"""
Various Utility functions used when creating/simulating the various backends.
"""

import matplotlib.pyplot as plt
import torch


def send_vars(variables, device):
    """
    Send a set of variables to a specific device. Used with PyTorch backends.

    :param variables:   Variables to send.
    :type variables:    List of torch.tensors
    :param device:      Device to send the variables to.
    :type device:       torch.device
    :return:            Variables sent to correct device.
    :rtype:             List of torch.tensors
    """
    for var in variables:
        var = var.to(device)
    torch.cuda.empty_cache()
    return variables

def spike_raster_plot(t,data,colors=None,offset=0) -> None:
    """
    Plot spike rasters of spiking data.

    :param t:   vector of timesteps.
    :type t:    List, np.ndarray, or torch.tensor
    :param data:    2D vector of spiking data. Each row corresponds to a different neuron.
    :type data:     np.ndarray or torch.tensor
    :param colors:  List of colors to plot each neuron, default is every neuron is blue.
    :type colors:   List of str, optional
    :param offset:  Constant vertical offset for all spikes, default is 0.
    :type offset:   Number, optional
    :return:        None
    :rtype:         N/A
    """
    if colors is None:
        colors = ['blue']
    if data.ndim > 1:
        for neuron in range(len(data)):
            spike_locs = []
            for step in range(len(t)):
                if data[neuron][step] > 0:
                    spike_locs.append(t[step])
            if len(colors) == 1:
                plt.eventplot(spike_locs,lineoffsets=neuron+1+offset, colors=colors[0],linelengths=0.8)
            else:
                plt.eventplot(spike_locs, lineoffsets=neuron + 1+offset, colors=colors[neuron], linelengths=0.8)
    else:
        spike_locs = []
        for step in range(len(t)):
            if data[step] > 0:
                spike_locs.append(t[step])
        plt.eventplot(spike_locs, lineoffsets=1+offset, colors=colors[0], linelengths=0.8)



"""
Neurons are the computational nodes of an SNS, and can be either spiking or non-spiking.
"""

"""
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
IMPORTS
"""

from typing import Dict, Any
import warnings
import numbers

from sns_toolbox.design.design_utilities import valid_color, set_text_color

"""
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
BASE CLASS
"""

class Neuron:
    """
    Parent class of all neurons.

    :param name:                    Name of this neuron preset, default is 'Neuron'.
    :type name:                     str, optional
    :param color:                   Background fill color for the neuron, default is 'white'.
    :type color:                    str, optional
    :param membrane_capacitance:    Neural membrane capacitance, default is 5.0. Units are nanofarads (nF).
    :type membrane_capacitance:     Number, optional
    :param membrane_conductance:    Neural membrane conductance, default is 1.0. Units are microsiemens (uS).
    :type membrane_conductance:     Number, optional
    :param bias:                    Internal bias current, default is 0.0. Units are nanoamps (nA).
    :type bias:                     Number, optional
    """
    def __init__(self, name: str = 'Neuron',
                 color: str = 'white',
                 membrane_capacitance: float = 5.0,
                 membrane_conductance: float = 1.0,
                 bias: float = 0.0) -> None:

        self.params: Dict[str, Any] = {}
        if valid_color(color):
            self.color = color
        else:
            warnings.warn('Specified color is not in the standard SVG set. Defaulting to white.')
            self.color = 'white'
        if isinstance(name,str):
            self.name = name
        else:
            raise TypeError('Neuron name must be a string')
        if isinstance(membrane_capacitance, numbers.Number):
            self.params['membrane_capacitance'] = membrane_capacitance
        else:
            raise TypeError('Membrane capacitance must be a number (int, float, double, etc.)')
        if isinstance(membrane_conductance, numbers.Number):
            self.params['membrane_conductance'] = membrane_conductance
        else:
            raise TypeError('Membrane conductance must be a number (int, float, double, etc.')
        if isinstance(bias,numbers.Number):
            self.params['bias'] = bias
        else:
            raise TypeError('Bias must be a number (int, float, double, etc.')
        self.params['spiking'] = False

"""
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
SPECIFIC MODELS
"""

class NonSpikingNeuron(Neuron):
    """
    Classic non-spiking neuron model, whose dynamics are as follows:
    membrane_capacitance*dU/dt = -membrane_conductance*u + bias current + synaptic current + external current.
    """
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

class SpikingNeuron(Neuron):
    """
    Generalized leaky integrate-and-fire neuron model, whose dynamics are as follows:
    membrane_capacitance*dU/dt = -membrane_conductance*u + bias current + synaptic current + external current;
    threshold_time_constant*dTheta/dt = -Theta + threshold_initial_value + threshold_proportionality_constant*u;
    if u > Theta, u->0.

    :param threshold_time_constant: Rate that the firing threshold moves to the baseline value, default is 5.0. Units
        are milliseconds (ms).
    :type threshold_time_constant:  Number, optional
    :param threshold_initial_value: Baseline value of the firing threshold, default is 1.0. Units are millivolts (mV).
    :type threshold_initial_value:  Number, optional
    :param threshold_proportionality_constant:  Constant which determines spiking behavior.
        In response to constant stimulus, negative values cause the firing rate to decrease, positive values cause the
        rate to increase, and zero causes the rate to remain constant. Default is 0.0.
    :param threshold_proportionality constant:  Number, optional
    """
    def __init__(self, threshold_time_constant: float = 5.0,
                 threshold_initial_value: float = 1.0,
                 threshold_proportionality_constant: float = 0.0,
                 **kwargs) -> None:

        super().__init__(**kwargs)
        self.params['threshold_time_constant'] = threshold_time_constant
        self.params['threshold_initial_value'] = threshold_initial_value
        self.params['threshold_proportionality_constant'] = threshold_proportionality_constant
        self.params['spiking'] = True



"""
Networks store neuron and synapse types, and the pattern of connections between them. While not capable of being
simulated directly, they are used as a template for compilation.
"""

"""
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
IMPORTS
"""

from typing import Dict, Any, List, Type, TypeVar
from numbers import Number
import copy
from graphviz import Digraph
import warnings
import numpy as np

from sns_toolbox.design.neurons import Neuron, NonSpikingNeuron, SpikingNeuron
from sns_toolbox.design.connections import Connection, NonSpikingSynapse, NonSpikingTransmissionSynapse, NonSpikingModulationSynapse
from sns_toolbox.design.design_utilities import valid_color, set_text_color

"""
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
BASE CLASS
"""

class Network:
    """
    Base class for all Networks.

    :param name: Name for this network, defaults to 'Network".
    :type name: str, optional
    :param R: Range of activity for the network, defaults to 20.0. Units are millivolts (mV).
    :type R: Number, optional
    """
    def __init__(self, name: str = 'Network', R: float = 20.0) -> None:
        self.params: Dict[str, Any] = {}
        if isinstance(name,str):
            self.params['name'] = name
        else:
            raise TypeError('Name must be a string')
        if isinstance(R,Number):
            if R > 0:
                self.params['R'] = R
            else:
                raise ValueError('R must be > 0')
        else:
            raise TypeError('R must be a number')
        self.inputs = []
        # self.inputConns = []
        self.populations = []
        self.outputs = []
        self.outputConns = []
        self.connections = []
        self.graph = Digraph(filename=(self.params['name']+'.gv'))

        # Compiler options
        self.params['spiking'] = False
        self.params['delay'] = False
        self.params['electrical'] = False
        self.params['electricalRectified'] = False

    def get_num_neurons(self) -> int:
        """
        Calculate the number of neurons in the network, including within populations.

        :return: The number of neurons in the network.
        :rtype: int
        """
        num_neurons = 0
        for pop in self.populations:
            num_neurons += pop['number']
        return num_neurons

    def get_num_connections(self) -> int:
        """
        Calculate the number of connections in the network. Does not account for individual synapses between neurons in
        populations.

        :return: The number of connections in the network.
        :rtype: int
        """
        return len(self.connections)

    def get_num_populations(self) -> int:
        """
        Get the number of populations in the network.

        :return: Number of populations.
        :rtype: int
        """
        return len(self.populations)

    def get_num_inputs(self) -> int:
        """
        Get the number of input nodes to the network. Vector inputs are treated as a single input node.

        :return: Number of inputs.
        :rtype: int
        """
        return len(self.inputs)

    def get_num_inputs_actual(self) -> int:
        """
        Calculate the number of individual inputs throughout the network. Vector inputs have one input per dimension.

        :return: Number of individual inputs.
        :rtype: int
        """
        index = 0
        for inp in range(self.get_num_inputs()):
            size = self.inputs[inp]['size']
            index += size
        return index

    def get_num_outputs(self) -> int:
        """
        Get the number of output nodes from the network. Vector outputs have one output node.

        :return: Number of outputs.
        :rtype: int
        """
        return len(self.outputs)

    def get_num_outputs_actual(self) -> int:
        """
        Calculate the number of individual outputs from the network. Vector outputs have one output per dimension.

        :return: Number of individual outputs.
        :rtype: int
        """
        index = 0
        for out in range(self.get_num_outputs()):
            source_pop = self.outputs[out]['source']
            num_source_neurons = self.populations[source_pop]['number']
            # for num in range(num_source_neurons):
            #     index += 1
            index += num_source_neurons
        return index

    def get_population_index(self, name: str) -> int:
        """
        Given a string, find the numerical index of the population corresponding to that name within the network.

        :param name: Name of the population to find.
        :type name: str
        :return: Index of the population.
        :rtype: int
        """
        if not isinstance(name,str):
            raise TypeError('Name must be a valid string')
        for index in range(len(self.populations)):
            if self.populations[index]['name'] == name:
                return index
        raise ValueError('Population not found by name \'%s\'' % str(name))

    def get_connection_index(self, name: str) -> int:
        """
        Given a string, find the numerical index of the connection corresponding to that name within the network.

        :param name: Name of the connection to find.
        :type name: str
        :return: Index of the connection.
        :rtype: int
        """
        if not isinstance(name,str):
            raise TypeError('Name must be a valid string')
        for index in range(len(self.populations)):
            if self.connections[index]['name'] == name:
                return index
        raise ValueError('Population not found by name \'%s\'' % str(name))

    def get_input_index(self, name: str) -> int:
        """
        Given a string, find the numerical index of the input node given by that name within the network.

        :param name: Name of the input node to find.
        :type name: str
        :return: Index of the input node.
        :rtype: int
        """
        if not isinstance(name,str):
            raise TypeError('Name must be a valid string')
        for index in range(len(self.inputs)):
            if self.inputs[index]['name'] == name:
                return index
        raise ValueError('Input not found by name \'%s\'' % str(name))

    def render_graph(self, imgFormat: str = 'png', view: bool = False) -> None:
        """
        Render an image of the network in the form of a directed graph (DG) using graphviz.

        :param imgFormat: File extension of the resulting image, default is 'png'.
        :type imgFormat: str, optional
        :param view: Boolean flag to view the image, default is 'False'.
        :type view: bool, optional
        :return: None
        :rtype: N/A
        """
        if not isinstance(view,bool):
            raise TypeError('View must be a boolean')
        self.graph.format = imgFormat
        self.graph.render(view=view,cleanup=True)

    # Construction
    def add_population(self, neuron_type: Neuron, shape, name: str = None, color=None, initial_value=None) -> None:
        """
        Add a neural population to the network.

        :param neuron_type:  Type of neuron to add.
        :type neuron_type: sns_toolbox.design.neurons.Neuron
        :param shape: The number of neurons in each dimension of the population.
        :type shape: np.ndarray or torch.tensor
        :param name: Name of the population, default is the original name.
        :type name: str, optional
        :param color: Color of the population in the rendered image, default is the original color.
        :type color: str, optional
        :param initial_value: Initial value of membrane voltage for each neuron in the population. Must be either a
            single value, or an array matching 'shape'. Default value is the original value.
        :type initial_value: class: 'numbers.Number', class: 'np.ndarray', or class: 'torch.tensor'
        :return: None
        :rtype: N/A
        """
        if not isinstance(neuron_type, Neuron):
            raise TypeError('Input type is not a neuron')
        if len(shape) > 1: # population is multidimensional
            total_num_neurons = shape[0] * shape[1]
        else:
            total_num_neurons = shape[0]
        if name is None:
            name = neuron_type.name
        elif not isinstance(name,str):
            raise TypeError('Name must be a string')
        if color is None:
            color = neuron_type.color
        elif not valid_color(color):
            warnings.warn('Specified color is not in the standard SVG set. Defaulting to base type color.')
            color = neuron_type.color
        font_color = set_text_color(color)
        if initial_value is None:
            if total_num_neurons > 1:
                if neuron_type is SpikingNeuron:
                    initial_value = np.linspace(0,neuron_type.params['threshold_initial_value'],num=total_num_neurons)
                else:
                    initial_value = np.linspace(0,self.params['R'],num=total_num_neurons)
            else:
                initial_value = 0.0
        self.populations.append({'type': copy.deepcopy(neuron_type),
                                 'number': int(total_num_neurons),
                                 'shape': shape,
                                 'name': name,
                                 'color': color,
                                 'initial_value': initial_value})
        if total_num_neurons > 1:
            if len(shape) > 1:
                self.graph.node(str(len(self.populations) - 1), name,
                                style='filled, rounded',
                                shape='box3d',  # Populations with multiple neurons are marked with an outline
                                fillcolor=color,
                                fontcolor=font_color)
            else:
                self.graph.node(str(len(self.populations)-1), name,
                                style='filled, rounded',
                                shape='rect',   # Populations with multiple neurons are marked with an outline
                                fillcolor=color,
                                fontcolor=font_color)
        else:
            self.graph.node(str(len(self.populations) - 1), name,
                            style='filled',
                            fillcolor=color,
                            fontcolor=font_color)

        if neuron_type.params['spiking']:
            self.params['spiking'] = True

    def add_neuron(self, neuron_type: Neuron, name=None, color=None, initial_value=0.0) -> None:
        """
        Add a neuron to the network. Note that this is just a special case of addPopulation, which makes a
        population of 1 neuron.

        :param neuron_type:  Type of neuron to add.
        :type neuron_type: sns_toolbox.design.neurons.Neuron
        :param name:        Name of the neuron, defaults to the original name.
        :type name:         str, optional
        :param color:       Color of the neuron in the visual render, defaults to the original color.
        :type color:        str, optional
        :param initial_value: Initial value of the membrane voltage, defaults to 0.0. Units are millivolts (mV).
        :type initial_value: Number
        :return:    None
        :rtype: N/A
        """
        self.add_population(neuron_type, shape=[1], name=name, color=color, initial_value=initial_value)

    def add_input(self, dest: Any, size: int = 1, name: str = 'Input', color='white') -> None:
        """
        Add an input source to the network.

        :param dest:        Destination this input connects to.
        :type dest:         int or string
        :param size:        Number of input elements, default is 1.
        :type size:         int, optional
        :param name:        Name of the input node, default is 'Input'.
        :type name:         str, optional
        :param color:       Color of the input node in the visual render, default is 'white'.
        :type color:        str, optional
        :return:    None
        :rtype:     N/A
        """

        if not isinstance(name,str):
            raise TypeError('Name must be a string')
        if not valid_color(color):
            warnings.warn('Specified color is not in the standard SVG set. Defaulting to white.')
            color = 'white'
        if not isinstance(size,int):
            raise TypeError('Input size must be an integer greater than zero')
        if size <= 0:
            raise ValueError('Input size must be greater than 0')
        if not isinstance(dest, int):
            if isinstance(dest, str):
                dest = self.get_population_index(dest)
            else:
                raise TypeError('Destination index must be an integer or name')
        if dest > (len(self.populations)-1):
            raise ValueError('Destination index is out of range')
        if (size > 1) and (self.populations[dest]['number'] != size):
            raise ValueError('Input vector must be either size 1 or same size as destination population')
        font_color = set_text_color(color)
        self.inputs.append({'name': name,
                            'size': size,
                            'destination': dest,
                            'color': color})
        if size == 1:
            self.graph.node('In'+str(len(self.inputs) - 1), name,
                            style='filled',
                            shape='invhouse',
                            fillcolor=color,
                            fontcolor=font_color)
        else:
            self.graph.node('In' + str(len(self.inputs) - 1), name,
                            style='filled,diagonals',
                            shape='invhouse',
                            fillcolor=color,
                            fontcolor=font_color)
        self.graph.edge('In'+str(len(self.inputs) - 1), str(dest))

    def add_output(self, source: Any, name: str = 'Output', spiking: bool = False, color: str = 'white') -> None:
        """
        Add an output node to the network.

        :param source:      Source this output is connected to.
        :type source:       int or str
        :param name:        Name of the node, defaults to 'Output'.
        :type name:         str, optional
        :param spiking:     Flag for if this node stores voltage or spikes, default is 'False'.
        :type spiking:      bool, optional
        :param color:       Color of the output in the visual render, default is 'white'.
        :type color:        str, optional
        :return: None
        :rtype: N/A
        """
        if isinstance(source,int):
            if source < 0:
                raise ValueError('Source must be an integer greater than or equal to 0')
        elif isinstance(source,str):
            source = self.get_population_index(source)
        else:
            raise TypeError('Source must be an integer greater than 0 or a name')
        if source > (len(self.populations)-1):
            raise ValueError('Source index is out of range')
        if not isinstance(name,str):
            raise TypeError('Name must be a string')
        if not isinstance(spiking,bool):
            raise TypeError('Spiking flag must be a boolean')
        if not valid_color(color):
            warnings.warn('Specified color is not in the standard SVG set. Defaulting to white.')
            color = 'white'
        font_color = set_text_color(color)
        self.outputs.append({'name': name,
                             'source': source,
                             'spiking': spiking,
                             'color': color})
        if spiking:
            if self.populations[source]['number'] == 1:
                self.graph.node('Out'+str(len(self.outputs) - 1), name,
                                style='filled',
                                shape='triangle',
                                fillcolor=color,
                                fontcolor=font_color)
            else:
                self.graph.node('Out'+str(len(self.outputs) - 1), name,
                                style='filled,diagonals',
                                shape='triangle',
                                fillcolor=color,
                                fontcolor=font_color)
        else:
            if self.populations[source]['number'] == 1:
                self.graph.node('Out' + str(len(self.outputs) - 1), name,
                                style='filled',
                                shape='house',
                                fillcolor=color,
                                fontcolor=font_color)
            else:
                self.graph.node('Out' + str(len(self.outputs) - 1), name,
                                style='filled,diagonals',
                                shape='house',
                                fillcolor=color,
                                fontcolor=font_color)

        self.graph.edge(str(source),'Out'+str(len(self.outputs)-1))

    def add_connection(self, connection_type: Connection, source: Any,
                       destination: Any, name: str = None, view_label: bool = False) -> None:
        """
        Add a synaptic connection between two populations in the network.

        :param connection_type: Type of connection to add.
        :type connection_type: sns_toolbox.design.connections.Connection
        :param source:      Index or name of source population in the network.
        :type source:       int or str
        :param destination: Index or name of destination population in the network.
        :type destination:  int or str
        :param name:        Name of synapse, default is the original name.
        :type name:         str, optional
        :param view_label:  Flag to render the name on the output graph, default is 'False'.
        :type view_label:   bool, optional
        :return: None
        :rtype: N/A
        """
        if not isinstance(connection_type, Connection):
            raise TypeError('Connection type must inherit from type Connection')
        if not isinstance(source,int):
            if isinstance(source,str):
                source = self.get_population_index(source)
            else:
                raise TypeError('Source index must be an integer or name')
        if not isinstance(destination,int):
            if isinstance(destination, str):
                destination = self.get_population_index(destination)
            else:
                raise TypeError('Destination index must be an integer or name')
        if not isinstance(view_label, bool):
            raise TypeError('view_label must be of type bool')
        if source > (len(self.populations)-1):
            raise ValueError('Source index is outside of network size')
        if source < 0:
            raise ValueError('Source index must be >= 0')
        if not isinstance(destination,int):
            if isinstance(destination,str):
                destination = self.get_population_index(destination)
            else:
                raise TypeError('Destination index must be an integer or name')
        if destination > (len(self.populations)-1):
            raise ValueError('Destination index is outside of network size')
        if destination < 0:
            raise ValueError('Destination index must be >= 0')
        if name is None:
            label = connection_type.params['name']
        else:
            if isinstance(name,str):
                label = name
            else:
                raise TypeError('Name must be a string')
        self.connections.append({'name': label,
                                 'source': source,
                                 'destination': destination,
                                 'params': connection_type.params,
                                 'type': copy.deepcopy(connection_type),
                                 'view': view_label})

        if connection_type.params['pattern']:   # Pattern connection
            if self.populations[source]['number'] == 1:
                raise TypeError('Pattern connections are not supported for source populations of size 1')
            elif self.populations[source]['shape'] != self.populations[destination]['shape']:
                raise TypeError('Pattern connections are not currently supported for populations of different shape')
            style = 'vee'
            direction = 'forward'
            # 2d populations
            if len(self.populations[source]['shape']) > 1:
                g_max = __kernel_connections_2d__(self.populations[source]['shape'],
                                                  connection_type.params['max_conductance'])
                self.connections[-1]['params']['max_conductance'] = g_max
                del_e = __kernel_connections_2d__(self.populations[source]['shape'],
                                                  connection_type.params['relative_reversal_potential'])
                self.connections[-1]['params']['relative_reversal_potential'] = del_e
                if connection_type.params['spiking']:
                    time_constant = __kernel_connections_2d__(self.populations[source]['shape'],
                                                              connection_type.params['synapticTimeConstant'])
                    self.connections[-1]['params']['synapticTimeConstant'] = time_constant
                    transmit_delay = __kernel_connections_2d__(self.populations[source]['shape'],
                                                               connection_type.params['transmissionDelay'])
                    self.connections[-1]['params']['transmissionDelay'] = transmit_delay
            # 1d populations
            else:
                g_max = __kernel_connections_1d__(self.populations[source]['number'],
                                                  connection_type.params['max_conductance'])
                self.connections[-1]['params']['max_conductance'] = g_max
                del_e = __kernel_connections_1d__(self.populations[source]['number'],
                                                  connection_type.params['relative_reversal_potential'])
                self.connections[-1]['params']['relative_reversal_potential'] = del_e
                if connection_type.params['spiking']:
                    time_constant = __kernel_connections_1d__(self.populations[source]['number'],
                                                              connection_type.params['synapticTimeConstant'])
                    self.connections[-1]['params']['synapticTimeConstant'] = time_constant
                    transmit_delay = __kernel_connections_1d__(self.populations[source]['number'],
                                                               connection_type.params['transmissionDelay'])
                    self.connections[-1]['params']['transmissionDelay'] = transmit_delay
        elif connection_type.params['electrical']:
            style = 'odiamond'
            if connection_type.params['rectified'] is False:
                direction = 'both'
                self.params['electrical'] = True
            else:
                direction = 'forward'
                self.params['electricalRectified'] = True
        else:   # Chemical synapse
            direction = 'forward'
            if connection_type.params['relative_reversal_potential'] > 0:
                style = 'invempty'
            elif connection_type.params['relative_reversal_potential'] < 0:
                style = 'dot'
            else:
                style = 'odot'

        if view_label:
            self.graph.edge(str(source),
                            str(destination), dir=direction, arrowhead=style, label=label, arrowtail=style)
        else:
            self.graph.edge(str(source),
                            str(destination), dir=direction, arrowhead=style, arrowtail=style)

        if connection_type.params['spiking']:
            self.params['spiking'] = True
            if connection_type.params['transmissionDelay'] > 0:
                self.params['delay'] = True

    def add_network(self, network: 'Network', color: str = None) -> None:
        """
        Add an existing topology of inputs, outputs, and populations to the network.

        :param network: Network to copy over.
        :type network:  sns_toolbox.design.networks.Network
        :param color:   Color to render nodes in the network, default is the original colors.
        :type color:    str, optional
        :return: None
        :rtype: N/A
        """
        if not isinstance(network, Network):
            raise TypeError('Network must be of type Network')

        num_inputs = len(self.inputs)
        num_populations = len(self.populations)
        num_outputs = len(self.outputs)

        if color is None:
            for population in network.populations:
                self.add_population(neuron_type=population['type'], shape=population['shape'],
                                    name=population['name'], color=population['color'])
            for inp in network.inputs:
                self.add_input(dest=inp['destination'] + num_populations, name=inp['name'], color=inp['color']
                               ,size=inp['size'])
            for out in network.outputs:
                self.add_output(source=out['source'] + num_populations, name=out['name'], color=out['color'],
                                spiking=out['spiking'])
            for connection in network.connections:
                self.add_connection(connection_type=connection['type'], source=connection['source'] + num_populations,
                                    destination=connection['destination']+num_populations, view_label=connection['view'])
        else:
            if not valid_color(color):
                warnings.warn('Specified color is not in the standard SVG set. Defaulting to white.')
                color = 'white'
            for population in network.populations:
                self.add_population(neuron_type=population['type'], shape=population['shape'],
                                    name=population['name'], color=color)
            for inp in network.inputs:
                self.add_input(dest=inp['destination'] + num_populations, name=inp['name'], color=color
                               ,size=inp['size'])
            for out in network.outputs:
                self.add_output(source=out['source'] + num_populations, name=out['name'], color=color,
                                spiking=out['spiking'])
            for connection in network.connections:
                self.add_connection(connection_type=connection['type'], source=connection['source'] + num_populations,
                                    destination=connection['destination']+num_populations, view_label=connection['view'])

    def copy(self):
        """
        Create a copy of the network.

        :return: A new network with the same properties as the original. Each can be edited without effecting the other.
        :rtype: sns_toolbox.design.networks.Network
        """
        new_net = Network(name=self.params['name'],R=self.params['R'])
        new_net.add_network(self)

        return new_net

"""
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
HELPER FUNCTIONS
"""

def __kernel_connections_1d__(pop_size,kernel):
    """
    Generate a connection matrix from a kernel vector and population size.

    :param pop_size: Number of neurons in the population.
    :type pop_size: int
    :param kernel: Kernel vector to apply.
    :type kernel: list, np.ndarray, or torch.tensor
    :return: Connection matrix.
    :rtype: np.ndarray
    """
    kernel_length = len(kernel)
    pad_amt = int((kernel_length-1)/2)
    connection_matrix = np.zeros([pop_size,pop_size])
    for row in range(pop_size):
        padded = np.zeros(pop_size + 2 * pad_amt)
        padded[row:row+kernel_length] = kernel
        connection_matrix[row,:] = padded[pad_amt:-pad_amt]
    return connection_matrix

def __kernel_connections_2d__(pop_shape,kernel):
    """
    Generate a connection matrix from a kernel matrix and population shape.

    :param pop_shape: Shape of the population.
    :type pop_shape: list, np.ndarray, or torch.tensor
    :param kernel: Kernel matrix to apply.
    :type kernel: np.ndarray or torch.Tensor
    :return: Connection matrix.
    :rtype: np.ndarray
    """
    kernel_rows = kernel.shape[0]
    kernel_cols = kernel.shape[1]
    num_kernel_dims = len(kernel.shape)
    pop_size = pop_shape[0]*pop_shape[1]
    pad_dims = []
    for dim in range(num_kernel_dims):
        pad_amt = int((kernel.shape[dim] - 1) / 2)
        pad_dims.append([pad_amt,pad_amt])
    source_matrix = np.zeros(pop_shape)
    connection_matrix = np.zeros([pop_size,pop_size])
    index = 0
    for row in range(pop_shape[0]):
        for col in range(pop_shape[1]):
            padded_matrix = np.pad(source_matrix, pad_dims)
            padded_matrix[row:row+kernel_rows,col:col+kernel_cols] = kernel
            pad_rows = pad_dims[0][0]
            pad_cols = pad_dims[1][0]
            if pad_cols == 0:
                subsection = padded_matrix[pad_rows:-pad_rows,:]
            elif pad_rows == 0:
                subsection = padded_matrix[:,pad_cols:-pad_cols]
            else:
                subsection = padded_matrix[pad_rows:-pad_rows,pad_cols:-pad_cols]
            connection_matrix[index,:] = subsection.flatten()
            index += 1
    return connection_matrix

"""
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
SPECIFIC MODELS
"""

class AdditionNetwork(Network):
    """
    Network which performs addition or subtraction of multiple inputs. Currently only supports non-spiking neurons.

    :param gains:       List of addition or subtraction weights.
    :type gains:        list, np.ndarray, or torch.tensor
    :param add_del_e:   Reversal potential of addition synapses, default is 100. Unit is millivolts (mV).
    :type add_del_e:    Number, optional
    :param sub_del_e:   Reversal potential of subtraction synapses, default is -40. Unit is millivolts (mV).
    :type sub_del_e:    Number, optional
    :param neuron_type: Neuron preset to use, default is sns_toolbox.design.neurons.NonSpikingNeuron.
    :type neuron_type:  sns_toolbox.design.neurons.NonSpikingNeuron, optional
    :param name:        Name of this network, default is 'Add'.
    :type name:         str, optional
    """
    def __init__(self,gains,add_del_e=100,sub_del_e=-40,neuron_type=NonSpikingNeuron(),name='Add',**kwargs):
        super().__init__(name=name,**kwargs)
        num_inputs = len(gains)
        self.add_neuron(neuron_type=neuron_type, name=name + 'Sum')
        for i in range(num_inputs):
            self.add_neuron(neuron_type, name=name + 'Src' + str(i))
            gain = gains[i]
            if gain > 0:
                conn = NonSpikingTransmissionSynapse(gain=gain, relative_reversal_potential=add_del_e, R=self.params['R'])
            else:
                conn = NonSpikingTransmissionSynapse(gain=gain, relative_reversal_potential=sub_del_e, R=self.params['R'])
            self.add_connection(conn, i + 1, name + 'Sum')

class MultiplicationNetwork(Network):
    """
    Network where the activity in neuron 0 is roughly multiplied by the activity in neuron 1. Currently only supports
    non-spiking neurons.

    :param neuron_type: Neuron preset to use, default is sns_toolbox.design.neurons.NonSpikingNeuron.
    :type neuron_type:  sns_toolbox.design.neurons.NonSpikingNeuron, optional
    :param name:        Name of this network, default is 'Multiply'.
    :type name:         str, optional
    """
    def __init__(self,neuron_type=NonSpikingNeuron(),name='Multiply',**kwargs):
        super().__init__(name=name,**kwargs)
        self.add_neuron(neuron_type, name=name + '0')
        self.add_neuron(neuron_type, name=name + '1')
        self.add_neuron(neuron_type, name=name + 'Inter')
        self.add_neuron(neuron_type, name=name + 'Result')

        transmit = NonSpikingTransmissionSynapse(gain=1.0,R=self.params['R'])
        conductance = -self.params['R']/-1.0
        modulate_special = NonSpikingSynapse(max_conductance=conductance, relative_reversal_potential=-1.0)

        self.add_connection(transmit, name + '0', name + 'Result')
        self.add_connection(modulate_special, name + '1', name + 'Inter')
        self.add_connection(modulate_special, name + 'Inter', name + 'Result')

class DivisionNetwork(Network):
    """
    Network where the activity in neuron 0 is roughly divided by the activity in neuron 1. Currently only supports
    non-spiking neurons.

    :param neuron_type: Neuron preset to use, default is sns_toolbox.design.neurons.NonSpikingNeuron.
    :type neuron_type:  sns_toolbox.design.neurons.NonSpikingNeuron, optional
    :param ratio:       Modulation ratio.
    :type ratio:        Number
    :param name:        Name of this network, default is 'Divide'.
    :type name:         str, optional
    """
    def __init__(self,gain,ratio,name='Divide',neuron_type=NonSpikingNeuron(),**kwargs):
        super().__init__(name=name,**kwargs)
        self.add_neuron(neuron_type, name=name + 'Transmit')
        self.add_neuron(neuron_type, name=name + 'Modulate')
        self.add_neuron(neuron_type, name=name + 'Results')

        transmission = NonSpikingTransmissionSynapse(gain,R=self.params['R'])
        self.add_connection(transmission, 0, 2)

        modulation = NonSpikingModulationSynapse(ratio)
        self.add_connection(modulation, 1, 2)

class DifferentiatorNetwork(Network):
    """
    Network where the activity of 'Uout' is the derivative of the activity in 'Uin'. Currently only supports
    non-spiking neurons.

    :param slew_rate:   Steepest signal this differentiator can handle, default is 1.0. Influences the difference in
        internal time constants.
    :type slew_rate:    Number, optional
    :param name:        Name of this network, default is 'Differentiate'.
    :type name:         str, optional
    :param tau_fast:    Time constant of the faster neurons, default is 1.0. Units are milliseconds (ms).
    :type tau_fast:     Number, optional
    """
    def __init__(self,slew_rate=1.0,name='Differentiate',tau_fast=1.0,**kwargs):
        super().__init__(name=name,**kwargs)
        fast_neuron_type = NonSpikingNeuron(membrane_capacitance=tau_fast,membrane_conductance=1.0)
        tau_slow = tau_fast + self.params['R']/slew_rate
        slow_neuron_type = NonSpikingNeuron(membrane_capacitance=tau_slow,membrane_conductance=1.0)
        add_synapse = NonSpikingTransmissionSynapse(gain=1.0,R=self.params['R'],relative_reversal_potential=40.0)
        sub_synapse = NonSpikingTransmissionSynapse(gain=-1.0,R=self.params['R'],relative_reversal_potential=-40.0)

        self.add_neuron(neuron_type=fast_neuron_type,name='Uin')
        self.add_neuron(neuron_type=fast_neuron_type,name='Ufast')
        self.add_neuron(neuron_type=slow_neuron_type,name='Uslow')
        self.add_neuron(neuron_type=fast_neuron_type,name='Uout')

        self.add_connection(add_synapse, 'Uin', 'Ufast')
        self.add_connection(add_synapse, 'Uin', 'Uslow')
        self.add_connection(add_synapse, 'Ufast', 'Uout')
        self.add_connection(sub_synapse, 'Uslow', 'Uout')

class IntegratorNetwork(Network):
    """
    Network where the activity of 'Uint' integrates the applied stimulus to itself over time. Currently only supports
    non-spiking neurons.

    :param integration_gain:            Gain of the integration, default is 1.0.
    :type integration_gain:             Number, optional
    :param relative_reversal_potential: Reversal potential of the synapses in the network, default is -40.0. Must be
        less than 0. Units are millivolts (mV).
    :type relative_reversal_potential:  Number, optional
    :param name:                        Name of this network, default is 'Integrator'.
    :type name:                         str, optional
    """
    def __init__(self, integration_gain=0.1, relative_reversal_potential=-40.0, name='Integrator', **kwargs):
        super().__init__(name=name,**kwargs)
        membrane_capacitance = 1/(2 * integration_gain)
        if relative_reversal_potential < 0.0:
            synaptic_conductance = -self.params['R']/relative_reversal_potential
        else:
            raise ValueError('Relative reversal potential (Delta E) must be less than zero for an integrator')

        neuron_type = NonSpikingNeuron(membrane_capacitance=membrane_capacitance,membrane_conductance=1.0,
                                       bias=self.params['R'])
        synapse_type = NonSpikingSynapse(max_conductance=synaptic_conductance,
                                         relative_reversal_potential=relative_reversal_potential)

        self.add_neuron(neuron_type,name='Uint')
        self.add_neuron(neuron_type)

        self.add_connection(synapse_type, 0, 1)
        self.add_connection(synapse_type, 1, 0)


# class AdaptationNetwork(Network):
#     def __init__(self,ratio=0.5,name='Adaptation',neuron_type=NonSpikingNeuron(),**kwargs):
#         super().__init__(name=name,**kwargs)
#         relative_reversal_potential = 2*(self.params['R']**2 + 1)/self.params['R']
#         conductance_fast_slow = self.params['R']/(relative_reversal_potential - self.params['R'])
#         conductance_slow_fast = (self.params['R']*(ratio-1)*(relative_reversal_potential - self.params['R'] +
#                                                              ratio*self.params['R']))/(ratio *
#                                                                                        relative_reversal_potential *
#                                                                                        (-relative_reversal_potential -
#                                                                                         ratio*self.params['R']))
#         fast_slow = NonSpikingSynapse(max_conductance=conductance_fast_slow,
#                                       relative_reversal_potential=relative_reversal_potential)
#         slow_fast = NonSpikingSynapse(max_conductance=conductance_slow_fast,
#                                       relative_reversal_potential=-relative_reversal_potential)
#
#         self.add_neuron(neuron_type,name='Uadapt')
#         self.add_neuron(neuron_type)
#
#         self.add_connection(fast_slow, 0, 1)
#         self.add_connection(slow_fast, 1, 0)


"""
Connections are the mechanisms for data transmission between neurons. They can either define an individual conductance-
based synapse, or a pattern of synapses which is tiled between two populations.
"""
import numpy as np

"""
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
IMPORTS
"""

from typing import Dict, Any
import numbers
import math

"""
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
BASE CLASSES
"""
class Connection:
    """
    Base class of all connections. Initializes a dictionary of parameters which is modified by classes which inherit
    from it.

    :param max_conductance: All connections have a maximum synaptic conductance. It can be a single value or a matrix,
        but it must be defined.
    :type max_conductance: Number, np.ndarray, or torch.tensor
    :param name: Name of this connection preset, defaults to 'Connection'.
    :type name: str, optional
    """
    def __init__(self, max_conductance, name: str = 'Connection'):
        self.params: Dict[str, Any] = {}
        self.params['spiking'] = False
        self.params['pattern'] = False
        self.params['electrical'] = False
        self.params['max_conductance'] = max_conductance
        if isinstance(name, str):
            self.params['name'] = name
        else:
            raise TypeError('Name should be a string')

class ElectricalSynapse(Connection):
    def __init__(self, conductance, name: str = 'Electrical Synapse', rect: bool = False) -> None:
        super().__init__(conductance, name)
        self.params['electrical'] = True
        self.params['rectified'] = rect

class NonSpikingConnection(Connection):
    """
    Base class of all non-spiking connections. Initializes a dictionary of parameters which is modified by classes
    which inherit from it.

    :param max_conductance: All connections have a maximum synaptic conductance. It can be a single value or a matrix,
        but it must be defined.
    :type max_conductance: Number, np.ndarray, or torch.tensor
    :param relative_reversal_potential: All chemical connections have a relative synaptic reversal potential. It can be
        a single value or a matrix, but it must be defined.
    :type relative_reversal_potential: Number, np.ndarray, or torch.tensor
    :param name: Name of this connection preset, defaults to 'Non-Spiking Connection'.
    :type name: str, optional
    """
    def __init__(self, max_conductance, relative_reversal_potential, name: str = 'Non-Spiking Connection') -> None:
        super().__init__(max_conductance, name)
        self.params['relative_reversal_potential'] = relative_reversal_potential
        self.params['spiking'] = False

class SpikingConnection(Connection):
    """
    Base class of all spiking connections. Initializes a dictionary of parameters which is modified by classes
    which inherit from it.

    :param max_conductance: All connections have a maximum synaptic conductance. It can be a single value or a matrix,
        but it must be defined.
    :type max_conductance: Number, np.ndarray, or torch.tensor
    :param relative_reversal_potential: All chemical connections have a relative synaptic reversal potential. It can be
        a single value or a matrix, but it must be defined.
    :type relative_reversal_potential: Number, np.ndarray, or torch.tensor
    :param name: Name of this connection preset, defaults to 'Spiking Connection'.
    :type name: str, optional
    """
    def __init__(self, max_conductance, relative_reversal_potential, time_constant, transmission_delay,
                 name: str = 'Spiking Connection') -> None:
        super().__init__(max_conductance, name)
        self.params['spiking'] = True
        self.params['relative_reversal_potential'] = relative_reversal_potential
        self.params['synapticTimeConstant'] = time_constant
        self.params['transmissionDelay'] = transmission_delay

class NonSpikingSynapse(NonSpikingConnection):
    """
    An individual non-spiking synapse, where the conductance is defined as Conductance = max_conductance * max(0,
    min(1, Upre/R)), and the synaptic current is i_syn = Conductance*(relative_reversal_potential - Upost).

    :param max_conductance: Maximum synaptic conductance, defaults to 1.0. Units are micro-siemens (uS).
    :type max_conductance: float, optional
    :param relative_reversal_potential: Synaptic reversal potential, defaults to 40.0. Units are millivolts (mV).
    :type relative_reversal_potential: float, optional
    """
    def __init__(self, max_conductance: float = 1.0,
                 relative_reversal_potential: float = 40.0,
                 **kwargs: Any) -> None:
        if isinstance(max_conductance, numbers.Number):
            if max_conductance <= 0:
                raise ValueError('max_conductance (gMax) must be greater than 0')
        else:
            raise TypeError('max_conductance (gMax) must be a number (int, float, double, etc.')
        if not isinstance(relative_reversal_potential, numbers.Number):
            raise TypeError('relative_reversal_potential (deltaEsyn) must be a number (int, float, double, etc.')
        super().__init__(max_conductance, relative_reversal_potential,**kwargs)  # Call to constructor of parent class
        self.params['pattern'] = False

class SpikingSynapse(SpikingConnection):
    """
    An individual spiking synapse, where the conductance is reset to max_conductance whenever the pre-synaptic
    neuron spikes, and otherwise decays to zero according to the time constant. Synaptic current
    is i_syn = Conductance*(relative_reversal_potential - Upost). Synaptic propagation can be delayed by a set number
    of timesteps.

    :param max_conductance: Maximum synaptic conductance, defaults to 1.0. Units are micro-siemens (uS).
    :type max_conductance: float, optional
    :param relative_reversal_potential: Synaptic reversal potential, defaults to 194.0. Units are millivolts (mV).
    :type relative_reversal_potential: float, optional
    :param time_constant: Time constant of synaptic decay, defaults to 1.0. Units are milliseconds (ms).
    :type time_constant: float, optional
    :param transmission_delay: Number of timesteps to delay synaptic activity, defaults to 0. Units are timesteps (dt).
    :type transmission_delay: int, optional
    """
    def __init__(self, max_conductance: float = 1.0,
                 relative_reversal_potential: float = 194.0,
                 time_constant: float = 1.0,
                 transmission_delay: int = 0,
                 **kwargs: Any) -> None:
        if isinstance(max_conductance, numbers.Number):
            if max_conductance <= 0:
                raise ValueError('max_conductance (gMax) must be greater than 0')
        else:
            raise TypeError('max_conductance (gMax) must be a number (int, float, double, etc.')
        if not isinstance(relative_reversal_potential, numbers.Number):
            raise TypeError('relative_reversal_potential (deltaEsyn) must be a number (int, float, double, etc.')
        super().__init__(max_conductance, relative_reversal_potential, time_constant, transmission_delay, **kwargs)  # Call to constructor of parent class
        self.params['pattern'] = False
        if isinstance(time_constant, numbers.Number):
            if time_constant > 0:
                self.params['synapticTimeConstant'] = time_constant
            else:
                raise ValueError('Synaptic time constant (tauS) must be greater than 0')
        else:
            raise TypeError('Synaptic time constant (tauS) must be a number (int, float, double, etc.) greater than 0')

        if isinstance(transmission_delay, int):
            if transmission_delay >= 0:
                self.params['synapticTransmissionDelay'] = transmission_delay
            else:
                raise ValueError('Synaptic transmission delay must be greater than or equal to zero')
        else:
            raise TypeError('Synaptic transmission delay must be an integer')

class NonSpikingPatternConnection(NonSpikingConnection):
    """
    A pattern of non-spiking synapses, with kernel matrices representing the maximum conductance and reversal potential
    of each synapse in the pattern.

    :param max_conductance_kernel: Kernel matrix of conductance values. Units are micro-siemens (uS).
    :type max_conductance_kernel: np.ndarray or torch.tensor
    :param relative_reversal_potential_kernel: Kernel matrix of reversal potential values. Units are millivolts (mV).
    :type relative_reversal_potential_kernel: np.ndarray or torch.tensor
    """
    def __init__(self, max_conductance_kernel, relative_reversal_potential_kernel, **kwargs: Any) -> None:
        if max_conductance_kernel.shape != relative_reversal_potential_kernel.shape:
            raise ValueError('Max Conductance and Relative Reversal Potential must be matrices of the same shape')
        if np.any(max_conductance_kernel <= 0):
            raise ValueError('Max Conductance values must be greater than zero')
        super().__init__(max_conductance_kernel, relative_reversal_potential_kernel, **kwargs)  # Call to constructor of parent class
        self.params['pattern'] = True

class SpikingPatternConnection(SpikingConnection):
    """
    A pattern of spiking synapses, with kernel matrices representing the maximum conductance, reversal potential, time
    constant, and transmission delay of each synapse in the pattern.

    :param max_conductance_kernel: Kernel matrix of conductance values. Units are micro-siemens (uS).
    :type max_conductance_kernel: np.ndarray or torch.tensor
    :param relative_reversal_potential_kernel: Kernel matrix of reversal potential values. Units are millivolts (mV).
    :type relative_reversal_potential_kernel: np.ndarray or torch.tensor
    :param time_constant_kernel: Kernel matrix of time constant values. Units are milliseconds (ms).
    :type time_constant_kernel: np.ndarray or torch.tensor
    :param transmission_delay_kernel: Kernel matrix of transmission delays. Units are timesteps (dt).
    :type transmission_delay_kernel: np.ndarray or torch.tensor
    """
    def __init__(self, max_conductance_kernel, relative_reversal_potential_kernel, time_constant_kernel,
                 transmission_delay_kernel, **kwargs: Any) -> None:
        if (max_conductance_kernel.shape != relative_reversal_potential_kernel.shape) or (
                max_conductance_kernel.shape != time_constant_kernel.shape) or (
                max_conductance_kernel.shape != transmission_delay_kernel.shape):
            raise ValueError('Max Conductance, Relative Reversal Potential, Time Constant, and Transmission Delay must be matrices of the same shape')
        if np.any(max_conductance_kernel <= 0):
            raise ValueError('Max Conductance values must be greater than zero')
        if np.any(time_constant_kernel <= 0):
            raise ValueError('Time constant values must be greater than 0 ms')
        if np.any(transmission_delay_kernel < 0):
            raise ValueError('Transmission delays must be non-negative')
        super().__init__(max_conductance_kernel, relative_reversal_potential_kernel, time_constant_kernel,
                         transmission_delay_kernel, **kwargs)  # Call to constructor of parent class
        self.params['pattern'] = True

"""
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
SPECIFIC MODELS
"""

class NonSpikingTransmissionSynapse(NonSpikingSynapse):
    """
    A non-spiking transmission synapse, where (given some integration_gain) the maximum conductance is
    max_conductance = (integration_gain*R)/(relative_reversal_potential - integration_gain*R).

    :param gain: Transmission integration gain, defaults to 1.0.
    :type gain: Number, optional
    :param name: Name of this synapse preset, defaults to 'Transmit'
    :type name: str, optional
    :param R: Range of neural voltage activity, defaults to 20.0. Units are millivolts (mV).
    :type R: Number
    """
    def __init__(self, gain: float = 1.0,
                 name: str = 'Transmit',
                 R: float = 20.0,
                 **kwargs) -> None:
        super().__init__(name=name, **kwargs)  # Call to constructor of parent class
        if isinstance(gain,numbers.Number): # Gain error handling
            if gain == 0:
                raise ValueError('Gain must be nonzero')
            elif math.copysign(1,gain) != math.copysign(1,self.params['relative_reversal_potential']):    # sign of integration_gain and DeltaE don't match
                raise ValueError('Gain of '+str(gain)+' and Relative Reversal Potential must have the same sign')
            else:
                try:
                    self.params['max_conductance'] = (gain*R)/(self.params['relative_reversal_potential']-gain*R)
                except ZeroDivisionError:
                    raise ValueError('Gain of '+str(gain)+' causes division by 0, decrease integration_gain or increase relative_reversal_potential')
                if self.params['max_conductance'] < 0:
                    raise ValueError('Gain of '+str(gain)+' causes max_conductance to be negative, decrease integration_gain or increase relative_reversal_potential_kernel')
        else:
            raise TypeError('Gain must be a number (int, float, double, etc.)')
        if isinstance(R, numbers.Number):   # R error handling
            if R <= 0:
                raise ValueError('R must be greater than zero')
        else:
            raise TypeError('R must be a number (int, float, double, etc.)')

class NonSpikingModulationSynapse(NonSpikingSynapse):
    """
    A non-spiking modulation synapse, where the relative_reversal_potential is set to 0.

    :param ratio: The desired ratio Upost/Upre when Upre is at max activity (R mV).
    :type ratio: Number
    :param name: Name of this synapse preset, defaults to 'Modulate'.
    :type name: str, optional
    """
    def __init__(self,ratio, name: str = 'Modulate', **kwargs) -> None:
        super().__init__(name=name,**kwargs)
        self.params['relative_reversal_potential'] = 0.0
        if isinstance(ratio, numbers.Number):   # ratio error handling
            if ratio <= 0:
                raise ValueError('Ratio must be greater than zero')
        else:
            raise ValueError('Ratio must be a number (int, float, double, etc.)')
        self.params['max_conductance'] = 1/ratio - 1

class SpikingTransmissionSynapse(SpikingSynapse):
    """
    A spiking version of the non-spiking transmission synapse.

    :param gain: Transmission frequency gain, defaults to 1.0.
    :type gain: Number, optional
    :param name: Name of this preset, defaults to 'Transmit'.
    :type name: str, optional
    :param max_frequency: Maximum spiking frequency, defaults to 10.0. Units are kilohertz (kHz).
    :type max_frequency: Number, optional
    :param non_linearity: A constant between 0 and 1 which limits the synaptic non-linearity. Values closer to 0 improve
        linearity.
    :type non_linearity: Number, optional
    """
    def __init__(self, gain: float = 1.0, name: str = 'Transmit', max_frequency: float = 10.0,
                 non_linearity: float = 0.1, **kwargs) -> None:
        if not isinstance(max_frequency, numbers.Number):
            raise TypeError('Max spiking frequency must be a number (int, float, double, etc.) greater than 0')
        elif max_frequency <= 0:
            raise ValueError('Max spiking frequency must be greater than 0')
        if isinstance(non_linearity, numbers.Number):
            if (non_linearity < 1.0) and (non_linearity > 0.0):
                time_constant = -1/(max_frequency * math.log(non_linearity))
            else:
                raise ValueError('Nonlinearity coefficient must be between 0 and 1')
        else:
            raise TypeError('Nonlinearity coefficient must be a number (int, float, double, etc.) between 0 and 1')
        super().__init__(time_constant=time_constant,name=name,**kwargs)
        if isinstance(gain,numbers.Number):
            if gain == 0:
                raise ValueError('Gain must be nonzero')
            elif math.copysign(1,gain) != math.copysign(1,self.params['relative_reversal_potential']):    # sign of integration_gain and DeltaE don't match
                raise ValueError('Gain of '+str(gain)+' and Relative Reversal Potential must have the same sign')
            else:
                try:
                    self.params['max_conductance'] = (gain*self.params['R'])/((self.params['relative_reversal_potential']-gain*-self.params['R'])*time_constant*max_frequency)
                except ZeroDivisionError:
                    raise ValueError('Gain of '+str(gain)+' causes division by 0, decrease integration_gain or increase relative_reversal_potential')
                if self.params['max_conductance'] < 0:
                    raise ValueError('Gain of '+str(gain)+' causes max_conductance to be negative, decrease integration_gain or increase relative_reversal_potential')
        else:
            raise TypeError('Gain of '+str(gain)+' must be a number (int, float, double, etc.)')

"""
Classes and functions for an easier pattern connection, just based on gains. Not fully implemented yet.
"""
# class NonSpikingGainPatternConnection(NonSpikingConnection):
#     def __init__(self,gain_matrix,
#                  name: str = 'Pattern',
#                  R: float = 20.0,
#                  positive_reversal_potential: float = 160.0,
#                  negative_reversal_potential: float = -80.0,
#                  wrap: bool = False):
#         """
#         Connection pattern between two neural populations (i.e. a kernel)
#         :param gain_matrix: Matrix (or vector for 1D kernels) of synaptic gains
#         :param name:    Name of this connection type
#         :param R:   Voltage range of neural activity (mV)
#         :param positive_reversal_potential: Reversal potential for connections with positive gain (mV)
#         :param negative_reversal_potential: Reversal potential for connections with negative gain (mV)
#         :param wrap:    Flag for if connections should wrap from one end of the population to the other
#         """
#         max_conductance = []
#         relative_reversal_potential = []
#         super().__init__(max_conductance,relative_reversal_potential,name=name)
#         self.params['wrap'] = wrap
#         self.params['R'] = R
#         self.params['pattern'] = True
#
#         self.positive_reversal_potential = positive_reversal_potential
#         self.negative_reversal_potential = negative_reversal_potential
#
#         if hasattr(gain_matrix[0],'__iter__'): # 2D kernel
#             for row in range(len(gain_matrix)):
#                 cond_values = []
#                 del_e_values = []
#                 for col in range(len(gain_matrix[0])):
#                     calc_max_conductance,calc_relative_reversal_potential = __calc_synaptic_parameters_from_gain__(gain_matrix[row][col],
#                                                                                                                    positive_reversal_potential,
#                                                                                                                    negative_reversal_potential,
#                                                                                                                    self.params['R'])
#                     cond_values.append(calc_max_conductance)
#                     del_e_values.append(calc_relative_reversal_potential)
#                 self.params['max_conductance_kernel'].append(cond_values)
#                 self.params['relative_reversal_potential_kernel'].append(del_e_values)
#         else:   # 1D kernel
#             for i in range(len(gain_matrix)):
#                 calc_max_conductance, calc_relative_reversal_potential = __calc_synaptic_parameters_from_gain__(gain_matrix[i],
#                                                                                                                 positive_reversal_potential,
#                                                                                                                 negative_reversal_potential,
#                                                                                                                 self.params['R'])
#                 self.params['max_conductance_kernel'].append(calc_max_conductance)
#                 self.params['relative_reversal_potential_kernel'].append(calc_relative_reversal_potential)

# def __calc_synaptic_parameters_from_gain__(gain, positive_reversal_potential, negative_reversal_potential, R):
#     if gain == 0.0:
#         return 0.0, 0.0
#     else:
#         if gain > 0.0:
#             relative_reversal_potential = positive_reversal_potential
#         else:
#             relative_reversal_potential = negative_reversal_potential
#         max_conductance = gain * R / (relative_reversal_potential - gain * R)
#
#         return max_conductance, relative_reversal_potential
#
# def __calc_spiking_synaptic_parameters_from_gain__(gain, positive_reversal_potential, negative_reversal_potential, R,
#                                                    time_constant, max_frequency):
#     if gain == 0.0:
#         return 0.0, 0.0
#     else:
#         if gain > 0.0:
#             relative_reversal_potential = positive_reversal_potential
#         else:
#             relative_reversal_potential = negative_reversal_potential
#         max_conductance = gain * R / ((relative_reversal_potential - gain * R)*time_constant*max_frequency)
#
#         return max_conductance, relative_reversal_potential



"""
Simulation backends for synthetic nervous system networks. Each of these are python-based, and are constructed using a
Network. They can then be run for a step, with the inputs being a vector of neural states and applied currents and the
output being the next step of neural states.
"""

"""
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
IMPORTS
"""

from typing import Any
import numpy as np
import torch
import sys
import warnings

from sns_toolbox.design.networks import Network
from sns_toolbox.design.neurons import SpikingNeuron

"""
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
BASE CLASS
"""

class __Backend__:
    """
    Base-level class for all simulation backends. Each will construct a representation of a given network using the
    desired backend technique, and take in (some form of) a vector of input states and applied currents, and compute
    the result for the next timestep.

    :param network:     Network to serve as a design template.
    :type network:      sns_toolbox.design.networks.Network
    :param dt:          Simulation time constant, default is 0.1. Units are milliseconds (ms).
    :type dt:           Number, optional
    :param debug:       When enabled, print debug information to the console. Default is 'False'.
    :type debug:        bool, optional
    :param substeps:    Number of simulation substeps before returning an output vector. Default is 1.
    :type substeps:     int, optional
    """
    def __init__(self, network: Network, dt: float = 0.1, debug: bool = False, substeps: int = 1) -> None:

        if substeps <= 0:
            raise ValueError('Substeps must be a positive integer')
        self.substeps = substeps
        self.network = network
        self.dt = dt
        self.debug = debug
        self.spiking = network.params['spiking']
        self.delay = network.params['delay']
        self.electrical = network.params['electrical']
        self.electrical_rectified = network.params['electricalRectified']

        if self.debug:
            print('#\nGETTING NET PARAMETERS\n#')
        self.__get_net_params__()

        if self.debug:
            print('#\nINITIALIZING VARIABLES AND PARAMETERS\n#')
        self.__initialize_vectors_and_matrices__()

        if self.debug:
            print('#\nSETTING NEURAL PARAMETERS\n#')
        self.__set_neurons__()

        if self.debug:
            print('#\nSETTING INPUT PARAMETERS\n#')
        self.__set_inputs__()

        if self.debug:
            print('#\nSETTING CONNECTION PARAMETERS\n#')
        self.__set_connections__()

        if self.debug:
            print('#\nCALCULATING TIME FACTORS\n#')
        self.__calculate_time_factors__()

        if self.spiking:
            if self.delay:
                if self.debug:
                    print('#\nINITIALIZING PROPAGATION DELAY\n#')
                self.__initialize_propagation_delay__()

        if self.debug:
            print('#\nSETTING OUTPUT PARAMETERS\n#')
        self.__set_outputs__()

        if self.debug:
            print('#\nALL FINAL PARAMETERS\n#')
            self.__debug_print__()
            print('#\nDONE BUILDING\n#')

    def __get_net_params__(self) -> None:
        """
        Get the main properties from the network. These are the number of populations, number of neurons, number of
        connections, number of inputs, number of outputs, and the network range of neural activity.

        :return: None
        :rtype: N/A
        """
        self.num_populations = self.network.get_num_populations()
        self.num_neurons = self.network.get_num_neurons()
        self.num_connections = self.network.get_num_connections()
        self.num_inputs = self.network.get_num_inputs()
        self.num_outputs = self.network.get_num_outputs()
        self.R = self.network.params['R']

        if self.debug:
            print('Number of Populations:')
            print(self.num_populations)
            print('Number of Neurons:')
            print(self.num_neurons)
            print('Number of Connections')
            print(self.num_connections)
            print('Number of Inputs:')
            print(self.num_inputs)
            print('Number of Outputs:')
            print('Network Voltage Range (mV):')
            print(self.R)

    def __initialize_vectors_and_matrices__(self) -> None:
        """
        Initialize all the vectors and matrices needed for all the neural states and parameters. That includes the
        following: U, ULast, Spikes, Cm, Gm, Ibias, Theta0, Theta, ThetaLast, m, TauTheta.

        :return:    None
        :rtype:     N/A
        """
        raise NotImplementedError

    def __set_neurons__(self) -> None:
        """
        Iterate over all populations in the network, and set the corresponding neural parameters for each neuron in the
        network: Cm, Gm, Ibias, ULast, U, Theta0, ThetaLast, Theta, TauTheta, m.

        :return:    None
        :rtype:     N/A
        """
        raise NotImplementedError

    def __set_inputs__(self) -> None:
        """
        Build the input connection matrix.

        :return:    None
        :rtype:     N/A
        """
        raise NotImplementedError

    def __set_connections__(self) -> None:
        """
        Build the synaptic parameter matrices. Interpret connectivity patterns between populations into individual
        synapses.

        :return:    None
        :rtype:     N/A
        """
        raise NotImplementedError

    def __calculate_time_factors__(self) -> None:
        """
        Precompute the time factors for the membrane voltage, firing threshold, and spiking synapses.

        :return:    None
        :rtype:     N/A
        """
        self.time_factor_membrane = self.dt / (self.c_m / self.g_m)
        if self.spiking:
            self.time_factor_threshold = self.dt / self.tau_theta
            self.time_factor_synapse = self.dt / self.tau_syn

    def __initialize_propagation_delay__(self) -> None:
        """
        Create a buffer sized to store enough spike data for the longest synaptic propagation delay.

        :return:    None
        :rtype:     N/A
        """
        raise NotImplementedError

    def __set_outputs__(self) -> None:
        """
        Build the output connectivity matrices for voltage and spike monitors. Generate separate output monitors for
        each neuron in a population.

        :return:    None
        :rtype:     N/A
        """
        raise NotImplementedError

    def __debug_print__(self) -> None:
        """
        Print the values for every vector/matrix which will be used in the forward computation.

        :return:    None
        :rtype:     N/A
        """
        print('Input Connectivity:')
        print(self.input_connectivity)
        print('g_max_non:')
        print(self.g_max_non)
        if self.spiking:
            print('GmaxSpike:')
            print(self.g_max_spike)
        print('del_e:')
        print(self.del_e)
        if self.electrical:
            print('Gelectrical:')
            print(self.g_electrical)
        if self.electrical_rectified:
            print('GelectricalRectified:')
            print(self.g_rectified)
        print('Output Voltage Connectivity')
        print(self.output_voltage_connectivity)
        if self.spiking:
            print('Output Spike Connectivity:')
            print(self.output_spike_connectivity)
        print('u:')
        print(self.u)
        print('u_last:')
        print(self.u_last)
        if self.spiking:
            print('theta_0:')
            print(self.theta_0)
            print('ThetaLast:')
            print(self.theta_last)
            print('Theta')
            print(self.theta)

    def forward(self, inputs) -> Any:
        """
        Compute the next neural states based on previous neural states. Handle substeps as well.

        :param inputs:  Input currents into the network.
        :type inputs:   np.ndarray or torch.tensor
        :return:        The neural states at the next step.
        :rtype:         np.ndarray or torch.tensor
        """
        for i in range(self.substeps):
            out = self.__forward_pass__(inputs)
        return out

    def __forward_pass__(self, inputs) -> Any:
        """
        Compute the next neural states based on previous neural states in the following steps:
        Ulast = U;
        ThetaLast = Theta;
        MappedInputs = cubic*inputs^3 + quadratic*inputs^2 + linear*inputs + offset;
        IApp = InputConnectivity X MappedInputs;
        GNon = max(0, min(GMaxNon*ULast/R, GMaxNon));
        GSpike = GSpike * (1-TimeFactorSynapse);
        GSyn = GNon + GSpike;
        ISyn = ColSum(GSyn*DelE) - ULast*ColSum(GSyn);
        U = ULast + TimeFactorMembrane*(-Gm*ULast + IBias + ISyn + IApp);
        Theta = ThetaLast + TimeFactorThreshold*(-ThetaLast + Theta0 + m*ULast);
        Spikes = Sign(min(0, Theta - U));
        SpikeBuffer = SpikeBuffer shifted down by 1;
        SpikeBuffer[first row] = Spikes;
        DelayedSpikeMatrix = SpikeBuffer[BufferSteps, BufferedNeurons];
        GSpike = max(GSpike, -DelayedSpikeMatrix*GMaxSpike);
        U = U * (Spikes + 1);
        Outputs = OutputVoltageConnectivity X U + OutputSpikeConnectivity X (-Spikes);
        MappedOutputs = cubic*Outputs^3 + quadratic*Outputs^2 + linear*Outputs + offset.

        :param inputs:  Input currents into the network.
        :type inputs:   np.ndarray or torch.tensor
        :return:        The neural states at the next step.
        :rtype:         np.ndarray or torch.tensor
        """
        raise NotImplementedError

"""
########################################################################################################################
NUMPY BACKEND

Simulating the network using numpy vectors and matrices.
Note that this is not sparse, so memory may explode for large networks
"""

class SNS_Numpy(__Backend__):
    """
    Simulation backend based in Numpy.

    :param network: Network which will be compiled to Numpy.
    :type network:  sns_toolbox.design.networks.Network
    """
    def __init__(self,network: Network,**kwargs):
        super().__init__(network,**kwargs)

    def __initialize_vectors_and_matrices__(self) -> None:

        self.u = np.zeros(self.num_neurons)
        self.u_last = np.zeros(self.num_neurons)
        self.c_m = np.zeros(self.num_neurons)
        self.g_m = np.zeros(self.num_neurons)
        self.i_b = np.zeros(self.num_neurons)
        if self.spiking:
            self.spikes = np.zeros(self.num_neurons)
            self.theta_0 = np.zeros(self.num_neurons)
            self.theta = np.zeros(self.num_neurons)
            self.theta_last = np.zeros(self.num_neurons)
            self.m = np.zeros(self.num_neurons)
            self.tau_theta = np.zeros(self.num_neurons)

        self.g_max_non = np.zeros([self.num_neurons, self.num_neurons])
        self.del_e = np.zeros([self.num_neurons, self.num_neurons])
        if self.spiking:
            self.g_max_spike = np.zeros([self.num_neurons, self.num_neurons])
            self.g_spike = np.zeros([self.num_neurons, self.num_neurons])
            self.tau_syn = np.zeros([self.num_neurons, self.num_neurons]) + 1
            if self.delay:
                self.spike_delays = np.zeros([self.num_neurons, self.num_neurons])
                self.spike_rows = []
                self.spike_cols = []
                self.buffer_steps = []
                self.buffer_nrns = []
                self.delayed_spikes = np.zeros([self.num_neurons, self.num_neurons])
        if self.electrical:
            self.g_electrical = np.zeros([self.num_neurons, self.num_neurons])
        if self.electrical_rectified:
            self.g_rectified = np.zeros([self.num_neurons, self.num_neurons])

        self.pops_and_nrns = []
        index = 0
        for pop in range(len(self.network.populations)):
            num_neurons = self.network.populations[pop]['number']  # find the number of neurons in the population
            self.pops_and_nrns.append([])
            for num in range(num_neurons):
                self.pops_and_nrns[pop].append(index)
                index += 1

    def __set_neurons__(self) -> None:

        index = 0
        for pop in range(len(self.network.populations)):
            num_neurons = self.network.populations[pop]['number']  # find the number of neurons in the population
            initial_value = self.network.populations[pop]['initial_value']
            for num in range(num_neurons):  # for each neuron, copy the parameters over
                self.c_m[index] = self.network.populations[pop]['type'].params['membrane_capacitance']
                self.g_m[index] = self.network.populations[pop]['type'].params['membrane_conductance']
                self.i_b[index] = self.network.populations[pop]['type'].params['bias']
                if hasattr(initial_value, '__iter__'):
                    self.u_last[index] = initial_value[num]
                elif initial_value is None:
                    self.u_last[index] = 0.0
                else:
                    self.u_last[index] = initial_value
                if self.spiking:
                    if isinstance(self.network.populations[pop]['type'], SpikingNeuron):  # if the neuron is spiking, copy more
                        self.theta_0[index] = self.network.populations[pop]['type'].params['threshold_initial_value']
                        self.m[index] = self.network.populations[pop]['type'].params['threshold_proportionality_constant']
                        self.tau_theta[index] = self.network.populations[pop]['type'].params['threshold_time_constant']
                    else:  # otherwise, set to the special values for NonSpiking
                        self.theta_0[index] = sys.float_info.max
                        self.m[index] = 0
                        self.tau_theta[index] = 1
                index += 1
        self.u = np.copy(self.u_last)
        if self.spiking:
            self.theta = np.copy(self.theta_0)
            self.theta_last = np.copy(self.theta_0)

    def __set_inputs__(self) -> None:

        self.input_connectivity = np.zeros([self.num_neurons, self.network.get_num_inputs_actual()])  # initialize connectivity matrix
        index = 0
        for inp in range(self.network.get_num_inputs()):  # iterate over the connections in the network
            size = self.network.inputs[inp]['size']
            dest_pop = self.network.inputs[inp]['destination']  # get the destination
            if size == 1:
                for dest in self.pops_and_nrns[dest_pop]:
                    self.input_connectivity[dest][inp] = 1.0  # set the weight in the correct source and destination
                index += 1
            else:
                for dest in self.pops_and_nrns[dest_pop]:
                    self.input_connectivity[dest][index] = 1.0
                    index += 1

    def __set_connections__(self) -> None:

        for syn in range(len(self.network.connections)):
            source_pop = self.network.connections[syn]['source']
            dest_pop = self.network.connections[syn]['destination']
            g_max = self.network.connections[syn]['params']['max_conductance']
            if self.network.connections[syn]['params']['electrical'] is False: # electrical connection
                del_e = self.network.connections[syn]['params']['relative_reversal_potential']

            if self.network.connections[syn]['params']['pattern']:  # pattern connection
                pop_size = len(self.pops_and_nrns[source_pop])
                source_index = self.pops_and_nrns[source_pop][0]
                dest_index = self.pops_and_nrns[dest_pop][0]
                if self.network.connections[syn]['params']['spiking']:
                    tau_s = self.network.connections[syn]['params']['synapticTimeConstant']
                    self.g_max_spike[dest_index:dest_index+pop_size,source_index:source_index+pop_size] = g_max
                    self.del_e[dest_index:dest_index+pop_size,source_index:source_index+pop_size] = del_e
                    self.tau_syn[dest_index:dest_index+pop_size,source_index:source_index+pop_size] = tau_s
                    if self.delay:
                        delay = self.network.connections[syn]['params']['synapticTransmissionDelay']
                        self.spike_delays[dest_index:dest_index+pop_size,source_index:source_index+pop_size] = delay

                        for source in self.pops_and_nrns[source_pop]:
                            for dest in self.pops_and_nrns[dest_pop]:
                                self.buffer_nrns.append(source)
                                self.buffer_steps.append(delay)
                                self.spike_rows.append(dest)
                                self.spike_cols.append(source)
                else:
                    self.g_max_non[dest_index:dest_index+pop_size,source_index:source_index+pop_size] = g_max
                    self.del_e[dest_index:dest_index+pop_size,source_index:source_index+pop_size] = del_e
            elif self.network.connections[syn]['params']['electrical']: # electrical connection
                for source in self.pops_and_nrns[source_pop]:
                    for dest in self.pops_and_nrns[dest_pop]:
                        if self.network.connections[syn]['params']['rectified']:    # rectified
                            self.g_rectified[dest][source] = g_max / len(self.pops_and_nrns[source_pop])
                        else:
                            self.g_electrical[dest][source] = g_max / len(self.pops_and_nrns[source_pop])
                            self.g_electrical[source][dest] = g_max / len(self.pops_and_nrns[source_pop])
            else:   # chemical connection
                if self.network.connections[syn]['params']['spiking']:  # spiking chemical synapse
                    tau_s = self.network.connections[syn]['params']['synapticTimeConstant']
                    if self.delay:
                        delay = self.network.connections[syn]['params']['synapticTransmissionDelay']
                    for source in self.pops_and_nrns[source_pop]:
                        for dest in self.pops_and_nrns[dest_pop]:
                            self.g_max_spike[dest][source] = g_max / len(self.pops_and_nrns[source_pop])
                            self.del_e[dest][source] = del_e
                            self.tau_syn[dest][source] = tau_s
                            if self.delay:
                                self.spike_delays[dest][source] = delay
                                self.buffer_nrns.append(source)
                                self.buffer_steps.append(delay)
                                self.spike_rows.append(dest)
                                self.spike_cols.append(source)
                else:   # nonspiking chemical synapse
                    for source in self.pops_and_nrns[source_pop]:
                        for dest in self.pops_and_nrns[dest_pop]:
                            self.g_max_non[dest][source] = g_max / len(self.pops_and_nrns[source_pop])
                            self.del_e[dest][source] = del_e

    def __initialize_propagation_delay__(self) -> None:
        buffer_length = int(np.max(self.spike_delays) + 1)
        self.spike_buffer = np.zeros([buffer_length, self.num_neurons])

    def __set_outputs__(self) -> None:

        outputs = []
        index = 0
        for out in range(len(self.network.outputs)):
            source_pop = self.network.outputs[out]['source']
            num_source_neurons = self.network.populations[source_pop]['number']
            outputs.append([])
            for num in range(num_source_neurons):
                outputs[out].append(index)
                index += 1
        self.num_outputs = index

        self.output_voltage_connectivity = np.zeros(
            [self.num_outputs, self.num_neurons])  # initialize connectivity matrix
        if self.spiking:
            self.output_spike_connectivity = np.copy(self.output_voltage_connectivity)
        self.outputs = np.zeros(self.num_outputs)
        for out in range(len(self.network.outputs)):  # iterate over the connections in the network
            source_pop = self.network.outputs[out]['source']  # get the source
            for i in range(len(self.pops_and_nrns[source_pop])):
                if self.network.outputs[out]['spiking']:
                    self.output_spike_connectivity[outputs[out][i]][
                        self.pops_and_nrns[source_pop][i]] = 1.0  # set the weight in the correct source and destination
                else:
                    self.output_voltage_connectivity[outputs[out][i]][
                        self.pops_and_nrns[source_pop][i]] = 1.0  # set the weight in the correct source and destination

    def __forward_pass__(self, inputs) -> Any:
        self.u_last = np.copy(self.u)
        i_app = np.matmul(self.input_connectivity, inputs)  # Apply external current sources to their destinations
        g_syn = np.maximum(0, np.minimum(self.g_max_non * self.u_last / self.R, self.g_max_non))
        if self.spiking:
            self.theta_last = np.copy(self.theta)
            self.g_spike = self.g_spike * (1 - self.time_factor_synapse)
            g_syn += self.g_spike

        i_syn = np.sum(g_syn * self.del_e, axis=1) - self.u_last * np.sum(g_syn, axis=1)
        if self.electrical:
            i_syn += (np.sum(self.g_electrical*self.u_last, axis=1) - self.u_last*np.sum(self.g_electrical, axis=1))
        if self.electrical_rectified:
            # create mask
            mask = np.subtract.outer(self.u_last, self.u_last).transpose() > 0
            masked_g = mask*self.g_rectified
            diag_masked = masked_g + masked_g.transpose() - np.diag(masked_g.diagonal())
            i_syn += np.sum(diag_masked*self.u_last, axis=1) - self.u_last*np.sum(diag_masked, axis=1)
        self.u = self.u_last + self.time_factor_membrane * (-self.g_m * self.u_last + self.i_b + i_syn + i_app)  # Update membrane potential
        if self.spiking:
            self.theta = self.theta_last + self.time_factor_threshold * (-self.theta_last + self.theta_0 + self.m * self.u_last)  # Update the firing thresholds
            self.spikes = np.sign(np.minimum(0, self.theta - self.u))  # Compute which neurons have spiked

            # New stuff with delay
            if self.delay:
                self.spike_buffer = np.roll(self.spike_buffer, 1, axis=0)   # Shift buffer entries down
                self.spike_buffer[0, :] = self.spikes    # Replace row 0 with the current spike data
                # Update a matrix with all of the appropriately delayed spike values
                self.delayed_spikes[self.spike_rows, self.spike_cols] = self.spike_buffer[self.buffer_steps, self.buffer_nrns]

                self.g_spike = np.maximum(self.g_spike, (-self.delayed_spikes) * self.g_max_spike)  # Update the conductance of connections which spiked
            else:
                self.g_spike = np.maximum(self.g_spike, (-self.spikes) * self.g_max_spike)  # Update the conductance of connections which spiked
            self.u = self.u * (self.spikes + 1)  # Reset the membrane voltages of neurons which spiked
        self.outputs = np.matmul(self.output_voltage_connectivity, self.u)
        if self.spiking:
            self.outputs += np.matmul(self.output_spike_connectivity, -self.spikes)

        return self.outputs



"""
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
PYTORCH DENSE

Simulating the network using GPU-compatible tensors.
Note that this is not sparse, so memory may explode for large networks
"""


class SNS_Torch(__Backend__):
    """
    Simulation backend based in PyTorch. In future versions different options will be toggled automatically, but for now
    are implemented as boolean flags.

    :param network: Network which will be compiled to PyTorch.
    :type network:  sns_toolbox.design.networks.Network
    :param device:  Device network will be stored on, default is 'cuda' (GPU).
    :type device:   str, optional
    """
    def __init__(self,network: Network,device: str = 'cuda',**kwargs):
        if device != 'cpu':
            if not torch.cuda.is_available():
                warnings.warn('CUDA Device Unavailable. Using CPU Instead')
                device = 'cpu'
        self.device = device
        super().__init__(network,**kwargs)

    def __initialize_vectors_and_matrices__(self) -> None:
        """
        Initialize all of the vectors and matrices needed for all of the neural states and parameters. That includes the
        following: U, ULast, Spikes, Cm, Gm, Ibias, Theta0, Theta, ThetaLast, m, TauTheta.
        :return:    None
        """
        self.u = torch.zeros(self.num_neurons,device=self.device)
        self.u_last = torch.zeros(self.num_neurons,device=self.device)
        self.c_m = torch.zeros(self.num_neurons,device=self.device)
        self.g_m = torch.zeros(self.num_neurons,device=self.device)
        self.i_b = torch.zeros(self.num_neurons,device=self.device)
        if self.spiking:
            self.spikes = torch.zeros(self.num_neurons, device=self.device)
            self.theta_0 = torch.zeros(self.num_neurons,device=self.device)
            self.theta = torch.zeros(self.num_neurons,device=self.device)
            self.theta_last = torch.zeros(self.num_neurons,device=self.device)
            self.m = torch.zeros(self.num_neurons,device=self.device)
            self.tau_theta = torch.zeros(self.num_neurons,device=self.device)

        self.g_max_non = torch.zeros([self.num_neurons, self.num_neurons],device=self.device)
        self.del_e = torch.zeros([self.num_neurons, self.num_neurons],device=self.device)
        if self.spiking:
            self.g_max_spike = torch.zeros([self.num_neurons, self.num_neurons], device=self.device)
            self.g_spike = torch.zeros([self.num_neurons, self.num_neurons], device=self.device)
            self.tau_syn = torch.ones([self.num_neurons, self.num_neurons],device=self.device)
            if self.delay:
                self.spike_delays = torch.zeros([self.num_neurons, self.num_neurons],device=self.device)
                self.spike_rows = []
                self.spike_cols = []
                self.buffer_steps = []
                self.buffer_nrns = []
                self.delayed_spikes = torch.zeros([self.num_neurons, self.num_neurons],device=self.device)
        if self.electrical:
            self.g_electrical = torch.zeros([self.num_neurons, self.num_neurons],device=self.device)
        if self.electrical_rectified:
            self.g_rectified = torch.zeros([self.num_neurons, self.num_neurons],device=self.device)

        self.pops_and_nrns = []
        index = 0
        for pop in range(len(self.network.populations)):
            num_neurons = self.network.populations[pop]['number']  # find the number of neurons in the population
            self.pops_and_nrns.append([])
            for num in range(num_neurons):
                self.pops_and_nrns[pop].append(index)
                index += 1

    def __set_neurons__(self) -> None:
        """
        Iterate over all populations in the network, and set the corresponding neural parameters for each neuron in the
        network: Cm, Gm, Ibias, ULast, U, Theta0, ThetaLast, Theta, TauTheta, m.
        :return:
        """
        index = 0
        for pop in range(len(self.network.populations)):
            num_neurons = self.network.populations[pop]['number']  # find the number of neurons in the population
            initial_value = self.network.populations[pop]['initial_value']
            for num in range(num_neurons):  # for each neuron, copy the parameters over
                self.c_m[index] = self.network.populations[pop]['type'].params['membrane_capacitance']
                self.g_m[index] = self.network.populations[pop]['type'].params['membrane_conductance']
                self.i_b[index] = self.network.populations[pop]['type'].params['bias']
                if hasattr(initial_value, '__iter__'):
                    self.u_last[index] = initial_value[num]
                elif initial_value is None:
                    self.u_last[index] = 0.0
                else:
                    self.u_last[index] = initial_value

                if self.spiking:
                    if isinstance(self.network.populations[pop]['type'], SpikingNeuron):  # if the neuron is spiking, copy more
                        self.theta_0[index] = self.network.populations[pop]['type'].params['threshold_initial_value']
                        self.m[index] = self.network.populations[pop]['type'].params['threshold_proportionality_constant']
                        self.tau_theta[index] = self.network.populations[pop]['type'].params['threshold_time_constant']
                    else:  # otherwise, set to the special values for NonSpiking
                        self.theta_0[index] = torch.finfo(self.theta_0[index].dtype).max
                        self.m[index] = 0
                        self.tau_theta[index] = 1
                index += 1
        self.u = self.u_last.clone()
        if self.spiking:
            self.theta = self.theta_0.clone()
            self.theta_last = self.theta_0.clone()

    def __set_inputs__(self) -> None:
        """
        Build the input connection matrix, and apply linear mapping coefficients.
        :return:    None
        """
        self.input_connectivity = torch.zeros([self.num_neurons, self.network.get_num_inputs_actual()],device=self.device)  # initialize connectivity matrix
        index = 0
        for inp in range(self.network.get_num_inputs()):  # iterate over the connections in the network
            size = self.network.inputs[inp]['size']
            dest_pop = self.network.inputs[inp]['destination']  # get the destination
            if size == 1:
                for dest in self.pops_and_nrns[dest_pop]:
                    self.input_connectivity[dest][inp] = 1.0  # set the weight in the correct source and destination
                index += 1
            else:
                for dest in self.pops_and_nrns[dest_pop]:
                    self.input_connectivity[dest][index] = 1.0
                    index += 1

    def __set_connections__(self) -> None:
        """
        Build the synaptic parameter matrices. Interpret connectivity patterns between populations into individual
        synapses.
        :return: None
        """
        for syn in range(len(self.network.connections)):
            source_pop = self.network.connections[syn]['source']
            dest_pop = self.network.connections[syn]['destination']
            g_max = self.network.connections[syn]['params']['max_conductance']
            if self.network.connections[syn]['params']['electrical'] is False:  # Chemical connection
                del_e = self.network.connections[syn]['params']['relative_reversal_potential']

            if self.network.connections[syn]['params']['pattern']:  # pattern connection
                pop_size = len(self.pops_and_nrns[source_pop])
                source_index = self.pops_and_nrns[source_pop][0]
                dest_index = self.pops_and_nrns[dest_pop][0]

                if self.network.connections[syn]['params']['spiking']:
                    tau_s = self.network.connections[syn]['params']['synapticTimeConstant']
                    self.g_max_spike[dest_index:dest_index + pop_size,source_index:source_index + pop_size] = torch.from_numpy(g_max)
                    self.del_e[dest_index:dest_index + pop_size,source_index:source_index + pop_size] = torch.from_numpy(del_e)
                    self.tau_syn[dest_index:dest_index+pop_size,source_index:source_index+pop_size] = torch.from_numpy(tau_s)
                    if self.delay:
                        delay = self.network.connections[syn]['params']['synapticTransmissionDelay']
                        self.spike_delays[dest_index:dest_index+pop_size,source_index:source_index+pop_size] = torch.from_numpy(delay)

                        for source in self.pops_and_nrns[source_pop]:
                            for dest in self.pops_and_nrns[dest_pop]:
                                self.buffer_nrns.append(source)
                                self.buffer_steps.append(delay)
                                self.spike_rows.append(dest)
                                self.spike_cols.append(source)
                else:
                    self.g_max_non[dest_index:dest_index+pop_size,source_index:source_index+pop_size] = torch.from_numpy(g_max)
                    self.del_e[dest_index:dest_index+pop_size,source_index:source_index+pop_size] = torch.from_numpy(del_e)
            elif self.network.connections[syn]['params']['electrical']:  # electrical connection
                for source in self.pops_and_nrns[source_pop]:
                    for dest in self.pops_and_nrns[dest_pop]:
                        if self.network.connections[syn]['params']['rectified']:  # rectified
                            self.g_rectified[dest][source] = g_max / len(self.pops_and_nrns[source_pop])
                        else:
                            self.g_electrical[dest][source] = g_max / len(self.pops_and_nrns[source_pop])
                            self.g_electrical[source][dest] = g_max / len(self.pops_and_nrns[source_pop])
            else:   # chemical connection
                if self.network.connections[syn]['params']['spiking']:  # spiking chemical synapse
                    tau_s = self.network.connections[syn]['params']['synapticTimeConstant']
                    if self.delay:
                        delay = self.network.connections[syn]['params']['synapticTransmissionDelay']
                    for source in self.pops_and_nrns[source_pop]:
                        for dest in self.pops_and_nrns[dest_pop]:
                            self.g_max_spike[dest][source] = g_max / len(self.pops_and_nrns[source_pop])
                            self.del_e[dest][source] = del_e
                            self.tau_syn[dest][source] = tau_s
                            if self.delay:
                                self.spike_delays[dest][source] = delay
                                self.buffer_nrns.append(source)
                                self.buffer_steps.append(delay)
                                self.spike_rows.append(dest)
                                self.spike_cols.append(source)
                else:   # nonspiking chemical synapse
                    for source in self.pops_and_nrns[source_pop]:
                        for dest in self.pops_and_nrns[dest_pop]:
                            self.g_max_non[dest][source] = g_max / len(self.pops_and_nrns[source_pop])
                            self.del_e[dest][source] = del_e

    def __initialize_propagation_delay__(self) -> None:
        """
        Create a buffer sized to store enough spike data for the longest synaptic propagation delay.
        :return: None
        """
        buffer_length = int(torch.max(self.spike_delays) + 1)
        self.spike_buffer = torch.zeros([buffer_length, self.num_neurons],device=self.device)

    def __set_outputs__(self) -> None:
        """
        Build the output connectivity matrices for voltage and spike monitors and apply linear maps. Generate separate
        output monitors for each neuron in a population.
        :return: None
        """
        outputs = []
        index = 0
        for out in range(len(self.network.outputs)):
            source_pop = self.network.outputs[out]['source']
            num_source_neurons = self.network.populations[source_pop]['number']
            outputs.append([])
            for num in range(num_source_neurons):
                outputs[out].append(index)
                index += 1
        self.num_outputs = index

        self.output_voltage_connectivity = torch.zeros(
            [self.num_outputs, self.num_neurons],device=self.device)  # initialize connectivity matrix
        if self.spiking:
            self.output_spike_connectivity = torch.clone(self.output_voltage_connectivity)
        self.outputs = torch.zeros(self.num_outputs, device=self.device)
        for out in range(len(self.network.outputs)):  # iterate over the connections in the network
            source_pop = self.network.outputs[out]['source']  # get the source
            for i in range(len(self.pops_and_nrns[source_pop])):
                if self.network.outputs[out]['spiking']:
                    self.output_spike_connectivity[outputs[out][i]][
                        self.pops_and_nrns[source_pop][i]] = 1.0  # set the weight in the correct source and destination
                    # self.out_linear[outputs[out][i]] = 1.0
                else:
                    self.output_voltage_connectivity[outputs[out][i]][
                        self.pops_and_nrns[source_pop][i]] = 1.0  # set the weight in the correct source and destination

    def __forward_pass__(self, inputs) -> Any:
        self.u_last = torch.clone(self.u)
        i_app = torch.matmul(self.input_connectivity, inputs)  # Apply external current sources to their destinations
        g_syn = torch.clamp(torch.minimum(self.g_max_non * self.u_last / self.R, self.g_max_non),min=0)
        if self.spiking:
            self.theta_last = torch.clone(self.theta)
            self.g_spike = self.g_spike * (1 - self.time_factor_synapse)
            g_syn += self.g_spike
        i_syn = torch.sum(g_syn * self.del_e, 1) - self.u_last * torch.sum(g_syn, 1)
        if self.electrical:
            i_syn += (torch.sum(self.g_electrical * self.u_last, 1) - self.u_last * torch.sum(self.g_electrical, 1))
        if self.electrical_rectified:
            # create mask
            mask = (self.u_last.reshape(-1,1)-self.u_last).transpose(0,1) > 0
            masked_g = mask * self.g_rectified
            diag_masked = masked_g + masked_g.transpose(0,1) - torch.diag(masked_g.diagonal())
            i_syn += torch.sum(diag_masked * self.u_last, 1) - self.u_last * torch.sum(diag_masked, 1)
        self.u = self.u_last + self.time_factor_membrane * (-self.g_m * self.u_last + self.i_b + i_syn + i_app)  # Update membrane potential
        if self.spiking:
            self.theta = self.theta_last + self.time_factor_threshold * (-self.theta_last + self.theta_0 + self.m * self.u_last)  # Update the firing thresholds
            self.spikes = torch.sign(torch.clamp(self.theta - self.u,max=0))  # Compute which neurons have spiked

            # New stuff with delay
            if self.delay:
                self.spike_buffer = torch.roll(self.spike_buffer, 1, 0)   # Shift buffer entries down
                self.spike_buffer[0, :] = self.spikes    # Replace row 0 with the current spike data
                # Update a matrix with all of the appropriately delayed spike values
                self.delayed_spikes[self.spike_rows, self.spike_cols] = self.spike_buffer[self.buffer_steps, self.buffer_nrns]

                self.g_spike = torch.maximum(self.g_spike, (-self.delayed_spikes) * self.g_max_spike)  # Update the conductance of connections which spiked
            else:
                self.g_spike = torch.maximum(self.g_spike, (-self.spikes) * self.g_max_spike)  # Update the conductance of connections which spiked
            self.u = self.u * (self.spikes + 1)  # Reset the membrane voltages of neurons which spiked
        self.outputs = torch.matmul(self.output_voltage_connectivity, self.u)
        if self.spiking:
            self.outputs += torch.matmul(self.output_spike_connectivity, -self.spikes)

        return self.outputs


"""
########################################################################################################################
PYTORCH SPARSE
"""
class SNS_Sparse(__Backend__):
    """
    Simulation backend based in PyTorch Sparse. In future versions different options will be toggled automatically,
    but for now are implemented as boolean flags.

    :param network: Network which will be compiled to PyTorch Sparse.
    :type network:  sns_toolbox.design.networks.Network
    :param device:  Device network will be stored on, default is 'cuda' (GPU).
    :type device:   str, optional
    """
    def __init__(self,network: Network,device: str = 'cuda',**kwargs):
        if device != 'cpu':
            if not torch.cuda.is_available():
                warnings.warn('CUDA Device Unavailable. Using CPU Instead')
                device = 'cpu'
        self.device = device
        super().__init__(network,**kwargs)

    def __initialize_vectors_and_matrices__(self) -> None:
        """
        Initialize all of the vectors and matrices needed for all of the neural states and parameters. That includes the
        following: U, ULast, Spikes, Cm, Gm, Ibias, Theta0, Theta, ThetaLast, m, TauTheta.
        :return:    None
        """
        self.u = torch.zeros(self.num_neurons,device=self.device)
        self.u_last = torch.zeros(self.num_neurons,device=self.device)
        self.c_m = torch.zeros(self.num_neurons,device=self.device)
        self.g_m = torch.zeros(self.num_neurons,device=self.device)
        self.i_b = torch.sparse_coo_tensor(size=(1,self.num_neurons),device=self.device)
        if self.spiking:
            self.spikes = torch.sparse_coo_tensor(size=(1, self.num_neurons), device=self.device)
            self.theta_0 = torch.zeros(self.num_neurons,device=self.device)
            self.theta = torch.zeros(self.num_neurons,device=self.device)
            self.theta_last = torch.zeros(self.num_neurons,device=self.device)
            self.m = torch.sparse_coo_tensor(size=(1,self.num_neurons),device=self.device)
            self.tau_theta = torch.zeros(self.num_neurons,device=self.device)

        self.g_max_non = torch.sparse_coo_tensor(size=(self.num_neurons,self.num_neurons),device=self.device)
        self.del_e = torch.sparse_coo_tensor(size=(self.num_neurons, self.num_neurons), device=self.device)
        if self.spiking:
            self.g_max_spike = torch.sparse_coo_tensor(size=(self.num_neurons,self.num_neurons),device=self.device)
            self.g_spike = torch.sparse_coo_tensor(size=(self.num_neurons,self.num_neurons),device=self.device)
            self.tau_syn = torch.ones([self.num_neurons, self.num_neurons],device=self.device)
            if self.delay:
                self.spike_delays = torch.sparse_coo_tensor(size=(self.num_neurons,self.num_neurons),device=self.device)
                self.spike_rows = []
                self.spike_cols = []
                self.buffer_steps = []
                self.buffer_nrns = []
                self.delayed_spikes = torch.sparse_coo_tensor(size=(self.num_neurons,self.num_neurons),device=self.device)
        if self.electrical:
            self.g_electrical = torch.sparse_coo_tensor(size=(self.num_neurons, self.num_neurons), device=self.device)
        if self.electrical_rectified:
            self.g_rectified = torch.sparse_coo_tensor(size=(self.num_neurons, self.num_neurons), device=self.device)

        self.pops_and_nrns = []
        index = 0
        for pop in range(len(self.network.populations)):
            num_neurons = self.network.populations[pop]['number']  # find the number of neurons in the population
            self.pops_and_nrns.append([])
            for num in range(num_neurons):
                self.pops_and_nrns[pop].append(index)
                index += 1

    def __set_neurons__(self) -> None:
        """
        Iterate over all populations in the network, and set the corresponding neural parameters for each neuron in the
        network: Cm, Gm, Ibias, ULast, U, Theta0, ThetaLast, Theta, TauTheta, m.
        :return:
        """
        index = 0
        for pop in range(len(self.network.populations)):
            num_neurons = self.network.populations[pop]['number']  # find the number of neurons in the population
            initial_value = self.network.populations[pop]['initial_value']
            for num in range(num_neurons):  # for each neuron, copy the parameters over
                self.c_m[index] = self.network.populations[pop]['type'].params['membrane_capacitance']
                self.g_m[index] = self.network.populations[pop]['type'].params['membrane_conductance']

                self.i_b = self.i_b.to_dense()
                self.i_b[0,index] = self.network.populations[pop]['type'].params['bias']
                self.i_b = self.i_b.to_sparse()

                if hasattr(initial_value, '__iter__'):
                    self.u_last[index] = initial_value[num]
                elif initial_value is None:
                    self.u_last[index] = 0.0
                else:
                    self.u_last[index] = initial_value

                if self.spiking:
                    if isinstance(self.network.populations[pop]['type'], SpikingNeuron):  # if the neuron is spiking, copy more
                        self.theta_0[index] = self.network.populations[pop]['type'].params['threshold_initial_value']

                        self.m = self.m.to_dense()
                        self.m[0,index] = self.network.populations[pop]['type'].params['threshold_proportionality_constant']
                        self.m = self.m.to_sparse()

                        self.tau_theta[index] = self.network.populations[pop]['type'].params['threshold_time_constant']
                    else:  # otherwise, set to the special values for NonSpiking
                        self.theta_0[index] = torch.finfo(self.theta_0[index].dtype).max

                        self.m = self.m.to_dense()
                        self.m[0,index] = 0
                        self.m = self.m.to_sparse()

                        self.tau_theta[index] = 1
                index += 1
        self.u = self.u_last.clone()
        if self.spiking:
            self.theta = self.theta_0.clone()
            self.theta_last = self.theta_0.clone()

    def __set_inputs__(self) -> None:
        """
        Build the input connection matrix, and apply linear mapping coefficients.
        :return:    None
        """
        self.input_connectivity = torch.sparse_coo_tensor(size=(self.num_neurons, self.network.get_num_inputs_actual()),device=self.device)  # initialize connectivity matrix
        index = 0
        for inp in range(self.network.get_num_inputs()):  # iterate over the connections in the network
            size = self.network.inputs[inp]['size']
            dest_pop = self.network.inputs[inp]['destination']  # get the destination

            self.input_connectivity = self.input_connectivity.to_dense()
            if size == 1:
                for dest in self.pops_and_nrns[dest_pop]:
                    self.input_connectivity[dest][inp] = 1.0  # set the weight in the correct source and destination
                index += 1
            else:
                for dest in self.pops_and_nrns[dest_pop]:
                    self.input_connectivity[dest][index] = 1.0
                    index += 1
            self.input_connectivity = self.input_connectivity.to_sparse()

    def __set_connections__(self) -> None:
        """
        Build the synaptic parameter matrices. Interpret connectivity patterns between populations into individual
        synapses.
        :return: None
        """
        for syn in range(len(self.network.connections)):
            source_pop = self.network.connections[syn]['source']
            dest_pop = self.network.connections[syn]['destination']
            g_max = self.network.connections[syn]['params']['max_conductance']
            if self.network.connections[syn]['params']['electrical'] is False:  # chemical connection
                del_e = self.network.connections[syn]['params']['relative_reversal_potential']

            if self.network.connections[syn]['params']['pattern']:  # pattern connection
                pop_size = len(self.pops_and_nrns[source_pop])
                source_index = self.pops_and_nrns[source_pop][0]
                dest_index = self.pops_and_nrns[dest_pop][0]

                if self.network.connections[syn]['params']['spiking']:
                    tau_s = self.network.connections[syn]['params']['synapticTimeConstant']

                    self.g_max_spike = self.g_max_spike.to_dense()
                    self.g_max_spike[dest_index:dest_index + pop_size,source_index:source_index + pop_size] = torch.from_numpy(g_max)
                    self.g_max_spike = self.g_max_spike.to_sparse()

                    self.del_e = self.del_e.to_dense()
                    self.del_e[dest_index:dest_index + pop_size,source_index:source_index + pop_size] = torch.from_numpy(del_e)
                    self.del_e = self.del_e.to_sparse()

                    self.tau_syn[dest_index:dest_index+pop_size,source_index:source_index+pop_size] = torch.from_numpy(tau_s)

                    if self.delay:
                        delay = self.network.connections[syn]['params']['synapticTransmissionDelay']
                        self.spike_delays = self.spike_delays.to_dense()
                        self.spike_delays[dest_index:dest_index+pop_size,source_index:source_index+pop_size] = torch.from_numpy(delay)
                        self.spike_delays = self.spike_delays.to_sparse()

                        for source in self.pops_and_nrns[source_pop]:
                            for dest in self.pops_and_nrns[dest_pop]:
                                self.buffer_nrns.append(source)
                                self.buffer_steps.append(delay)
                                self.spike_rows.append(dest)
                                self.spike_cols.append(source)
                else:
                    self.g_max_non = self.g_max_non.to_dense()
                    self.g_max_non[dest_index:dest_index+pop_size,source_index:source_index+pop_size] = torch.from_numpy(g_max)
                    self.g_max_non = self.g_max_non.to_sparse()

                    self.del_e = self.del_e.to_dense()
                    self.del_e[dest_index:dest_index+pop_size,source_index:source_index+pop_size] = torch.from_numpy(del_e)
                    self.del_e = self.del_e.to_sparse()
            elif self.network.connections[syn]['params']['electrical']:  # electrical connection
                for source in self.pops_and_nrns[source_pop]:
                    for dest in self.pops_and_nrns[dest_pop]:
                        if self.network.connections[syn]['params']['rectified']:  # rectified
                            self.g_rectified = self.g_rectified.to_dense()
                            self.g_rectified[dest][source] = g_max / len(self.pops_and_nrns[source_pop])
                            self.g_rectified = self.g_rectified.to_sparse()
                        else:
                            self.g_electrical = self.g_electrical.to_dense()
                            self.g_electrical[dest][source] = g_max / len(self.pops_and_nrns[source_pop])
                            self.g_electrical[source][dest] = g_max / len(self.pops_and_nrns[source_pop])
                            self.g_electrical = self.g_electrical.to_sparse()
            else:   # chemical connections
                if self.network.connections[syn]['params']['spiking']:  # spiking chemical synapse
                    tau_s = self.network.connections[syn]['params']['synapticTimeConstant']
                    if self.delay:
                        delay = self.network.connections[syn]['params']['synapticTransmissionDelay']
                    for source in self.pops_and_nrns[source_pop]:
                        for dest in self.pops_and_nrns[dest_pop]:
                            self.g_max_spike = self.g_max_spike.to_dense()
                            self.g_max_spike[dest][source] = g_max / len(self.pops_and_nrns[source_pop])
                            self.g_max_spike = self.g_max_spike.to_sparse()

                            self.del_e = self.del_e.to_dense()
                            self.del_e[dest][source] = del_e
                            self.del_e = self.del_e.to_sparse()

                            self.tau_syn[dest][source] = tau_s

                            if self.delay:
                                self.spike_delays = self.spike_delays.to_dense()
                                self.spike_delays[dest][source] = delay
                                self.spike_delays = self.spike_delays.to_sparse()

                                self.buffer_nrns.append(source)
                                self.buffer_steps.append(delay)
                                self.spike_rows.append(dest)
                                self.spike_cols.append(source)
                else:   # non-spiking chemical synapse
                    for source in self.pops_and_nrns[source_pop]:
                        for dest in self.pops_and_nrns[dest_pop]:
                            self.g_max_non = self.g_max_non.to_dense()
                            self.g_max_non[dest][source] = g_max / len(self.pops_and_nrns[source_pop])
                            self.g_max_non = self.g_max_non.to_sparse()

                            self.del_e = self.del_e.to_dense()
                            self.del_e[dest][source] = del_e
                            self.del_e = self.del_e.to_sparse()

    def __initialize_propagation_delay__(self) -> None:
        """
        Create a buffer sized to store enough spike data for the longest synaptic propagation delay.
        :return: None
        """
        self.spike_delays = self.spike_delays.to_dense()
        buffer_length = int(torch.max(self.spike_delays) + 1)
        self.spike_delays = self.spike_delays.to_sparse()

        self.spike_buffer = torch.sparse_coo_tensor(size=(buffer_length,self.num_neurons),device=self.device)

    def __set_outputs__(self) -> None:
        """
        Build the output connectivity matrices for voltage and spike monitors and apply linear maps. Generate separate
        output monitors for each neuron in a population.
        :return: None
        """
        outputs = []
        index = 0
        for out in range(len(self.network.outputs)):
            source_pop = self.network.outputs[out]['source']
            num_source_neurons = self.network.populations[source_pop]['number']
            outputs.append([])
            for num in range(num_source_neurons):
                outputs[out].append(index)
                index += 1
        self.num_outputs = index

        self.output_voltage_connectivity = torch.sparse_coo_tensor(size=(self.num_outputs, self.num_neurons),device=self.device)  # initialize connectivity matrix
        if self.spiking:
            self.output_spike_connectivity = torch.clone(self.output_voltage_connectivity)
        self.outputs = torch.sparse_coo_tensor(size=(1, self.num_outputs), device=self.device)

        for out in range(len(self.network.outputs)):  # iterate over the connections in the network
            source_pop = self.network.outputs[out]['source']  # get the source
            for i in range(len(self.pops_and_nrns[source_pop])):
                if self.network.outputs[out]['spiking']:
                    self.output_spike_connectivity = self.output_spike_connectivity.to_dense()
                    self.output_spike_connectivity[outputs[out][i]][
                        self.pops_and_nrns[source_pop][i]] = 1.0  # set the weight in the correct source and destination
                    self.output_spike_connectivity = self.output_spike_connectivity.to_sparse()
                else:
                    self.output_voltage_connectivity = self.output_voltage_connectivity.to_dense()
                    self.output_voltage_connectivity[outputs[out][i]][
                        self.pops_and_nrns[source_pop][i]] = 1.0  # set the weight in the correct source and destination
                    self.output_voltage_connectivity = self.output_voltage_connectivity.to_sparse()

    def __forward_pass__(self, inputs) -> Any:
        self.u_last = torch.clone(self.u)

        i_app = torch.matmul(self.input_connectivity, inputs)  # Apply external current sources to their destinations
        i_app = i_app.to_sparse()

        g_syn = torch.clamp(torch.minimum(self.g_max_non.to_dense() * self.u_last / self.R, self.g_max_non.to_dense()),min=0)
        g_syn = g_syn.to_sparse()

        if self.spiking:
            self.theta_last = torch.clone(self.theta)
            self.g_spike = self.g_spike.to_dense() * (1 - self.time_factor_synapse)
            self.g_spike = self.g_spike.to_sparse()

            g_syn += self.g_spike

        if g_syn._nnz() > 0:
            i_syn = torch.sparse.sum(g_syn * self.del_e, 1) - (self.u_last * torch.sum(g_syn.to_dense(), 1)).to_sparse()
        else:
            i_syn = torch.sparse.sum(g_syn * self.del_e) - self.u_last * torch.sparse.sum(g_syn)
        if self.electrical:
            i_syn += (torch.sum(self.g_electrical.to_dense() * self.u_last, 1).to_sparse() -
                      (self.u_last * torch.sum(self.g_electrical.to_dense(), 1)).to_sparse())
        if self.electrical_rectified:
            # create mask
            mask = (self.u_last.reshape(-1, 1) - self.u_last).transpose(0, 1) > 0
            masked_g = mask * self.g_rectified.to_dense()
            diag_masked = masked_g + masked_g.transpose(0, 1) - torch.diag(masked_g.diagonal())
            i_syn += torch.sum(diag_masked * self.u_last, 1).to_sparse() - (self.u_last * torch.sum(diag_masked, 1)).to_sparse()
        self.u = self.u_last + self.time_factor_membrane * (-self.g_m * self.u_last + (self.i_b.to_dense())[0,:] + i_syn + i_app)  # Update membrane potential
        if self.spiking:
            self.theta = self.theta_last + self.time_factor_threshold * (-self.theta_last + self.theta_0 + (self.m.to_dense())[0,:] * self.u_last)  # Update the firing thresholds

            self.spikes = torch.sign(torch.clamp(self.theta - self.u,max=0))  # Compute which neurons have spiked
            self.spikes = self.spikes.to_sparse()

            if self.delay:
                # New stuff with delay
                self.spike_buffer = self.spike_buffer.to_dense()
                self.spike_buffer = torch.roll(self.spike_buffer, 1, 0)   # Shift buffer entries down
                self.spike_buffer[0, :] = self.spikes.to_dense()    # Replace row 0 with the current spike data
                self.spike_buffer = self.spike_buffer.to_sparse()

                # Update a matrix with all of the appropriately delayed spike values
                self.delayed_spikes = self.delayed_spikes.to_dense()
                self.delayed_spikes[self.spike_rows, self.spike_cols] = (self.spike_buffer.to_dense())[self.buffer_steps, self.buffer_nrns]
                self.delayed_spikes = self.delayed_spikes.to_sparse()

                self.g_spike = torch.maximum(self.g_spike.to_dense(), ((-self.delayed_spikes) * self.g_max_spike).to_dense())  # Update the conductance of connections which spiked
            else:
                self.g_spike = torch.maximum(self.g_spike.to_dense(), (-self.spikes.to_dense()) * self.g_max_spike.to_dense())  # Update the conductance of connections which spiked
            self.g_spike = self.g_spike.to_sparse()
            self.u = self.u * (self.spikes.to_dense() + 1)  # Reset the membrane voltages of neurons which spiked
        self.outputs = torch.matmul(self.output_voltage_connectivity, self.u)
        if self.spiking:
            self.outputs += torch.matmul(self.output_spike_connectivity, -self.spikes.to_dense())

        return self.outputs


"""
########################################################################################################################
MANUAL BACKEND

Simulating the network using numpy vectors and matrices.
Note that this is not sparse, so memory may explode for large networks
"""
class SNS_Manual(__Backend__):
    """
    Simulation backend based in Numpy, but computes neural and synapses states in loops. Primarily for comparison, but
    could be useful for giant networks that don't fit in memory. In future versions different options will be toggled
    automatically, but for now are implemented as boolean flags.

    :param network: Network which will be compiled.
    :type network:  sns_toolbox.design.networks.Network
    """
    def __init__(self,network: Network,**kwargs):
        super().__init__(network,**kwargs)

    def __initialize_vectors_and_matrices__(self) -> None:
        """
        Initialize all of the vectors and matrices needed for all of the neural states and parameters. That includes the
        following: U, ULast, Spikes, Cm, Gm, Ibias, Theta0, Theta, ThetaLast, m, TauTheta.
        :return:    None
        """
        self.u = np.zeros(self.num_neurons)
        self.u_last = np.zeros(self.num_neurons)
        self.c_m = np.zeros(self.num_neurons)
        self.g_m = np.zeros(self.num_neurons)
        self.i_b = np.zeros(self.num_neurons)
        if self.spiking:
            self.spikes = np.zeros(self.num_neurons)
            self.theta_0 = np.zeros(self.num_neurons)
            self.theta = np.zeros(self.num_neurons)
            self.theta_last = np.zeros(self.num_neurons)
            self.m = np.zeros(self.num_neurons)
            self.tau_theta = np.zeros(self.num_neurons)

        self.incoming_synapses = []
        for i in range(self.num_neurons):
            self.incoming_synapses.append([])

        self.pops_and_nrns = []
        index = 0
        for pop in range(len(self.network.populations)):
            num_neurons = self.network.populations[pop]['number']  # find the number of neurons in the population
            self.pops_and_nrns.append([])
            for num in range(num_neurons):
                self.pops_and_nrns[pop].append(index)
                index += 1

    def __set_neurons__(self) -> None:
        """
        Iterate over all populations in the network, and set the corresponding neural parameters for each neuron in the
        network: Cm, Gm, Ibias, ULast, U, Theta0, ThetaLast, Theta, TauTheta, m.
        :return:
        """
        index = 0
        for pop in range(len(self.network.populations)):
            num_neurons = self.network.populations[pop]['number']  # find the number of neurons in the population
            initial_value = self.network.populations[pop]['initial_value']
            for num in range(num_neurons):  # for each neuron, copy the parameters over
                self.c_m[index] = self.network.populations[pop]['type'].params['membrane_capacitance']
                self.g_m[index] = self.network.populations[pop]['type'].params['membrane_conductance']
                self.i_b[index] = self.network.populations[pop]['type'].params['bias']
                if hasattr(initial_value, '__iter__'):
                    self.u_last[index] = initial_value[num]
                elif initial_value is None:
                    self.u_last[index] = 0.0
                else:
                    self.u_last[index] = initial_value
                if self.spiking:
                    if isinstance(self.network.populations[pop]['type'], SpikingNeuron):  # if the neuron is spiking, copy more
                        self.theta_0[index] = self.network.populations[pop]['type'].params['threshold_initial_value']
                        self.m[index] = self.network.populations[pop]['type'].params['threshold_proportionality_constant']
                        self.tau_theta[index] = self.network.populations[pop]['type'].params['threshold_time_constant']
                    else:  # otherwise, set to the special values for NonSpiking
                        self.theta_0[index] = sys.float_info.max
                        self.m[index] = 0
                        self.tau_theta[index] = 1
                index += 1
        self.u = np.copy(self.u_last)
        if self.spiking:
            self.theta = np.copy(self.theta_0)
            self.theta_last = np.copy(self.theta_0)

    def __set_inputs__(self) -> None:
        """
        Build the input connection matrix, and apply linear mapping coefficients.
        :return:    None
        """
        self.input_connectivity = np.zeros([self.num_neurons, self.network.get_num_inputs_actual()])  # initialize connectivity matrix
        index = 0
        for inp in range(self.network.get_num_inputs()):  # iterate over the connections in the network
            size = self.network.inputs[inp]['size']
            dest_pop = self.network.inputs[inp]['destination']  # get the destination
            if size == 1:
                for dest in self.pops_and_nrns[dest_pop]:
                    self.input_connectivity[dest][inp] = 1.0  # set the weight in the correct source and destination
                index += 1
            else:
                for dest in self.pops_and_nrns[dest_pop]:
                    self.input_connectivity[dest][index] = 1.0
                    index += 1

    def __set_connections__(self) -> None:
        """
        Build the synaptic parameter matrices. Interpret connectivity patterns between populations into individual
        synapses.
        :return: None
        """
        for syn in range(len(self.network.connections)):
            source_pop = self.network.connections[syn]['source']
            dest_pop = self.network.connections[syn]['destination']
            g_max = self.network.connections[syn]['params']['max_conductance']
            if self.network.connections[syn]['params']['electrical'] is False:  # electrical connection
                del_e = self.network.connections[syn]['params']['relative_reversal_potential']

            if self.network.connections[syn]['params']['pattern']:  # pattern connection
                pop_size = len(self.pops_and_nrns[source_pop])
                source_index = self.pops_and_nrns[source_pop][0]
                dest_index = self.pops_and_nrns[dest_pop][0]
                if self.network.connections[syn]['params']['spiking']:
                    tau_s = self.network.connections[syn]['params']['synapticTimeConstant']
                    if self.delay:
                        delay = self.network.connections[syn]['params']['synapticTransmissionDelay']

                    for dest in range(pop_size):
                        for source in range(pop_size):
                            g_syn = g_max[dest,source]
                            rev = del_e[dest,source]
                            time_factor_syn = self.dt/tau_s[dest,source]
                            if self.delay:
                                buffer = np.zeros(delay[dest, source])
                                self.incoming_synapses[dest+dest_index].append([source+source_index, True, False, g_syn, rev, 0, time_factor_syn, buffer])
                            else:
                                self.incoming_synapses[dest + dest_index].append([source + source_index, True, False, g_syn, rev, 0, time_factor_syn])
                else:
                    for dest in range(pop_size):
                        for source in range(pop_size):
                            g_syn = g_max[dest, source]
                            rev = del_e[dest, source]

                            self.incoming_synapses[dest + dest_index].append([source + source_index, False, False, g_syn, rev, 0])
            elif self.network.connections[syn]['params']['electrical']:  # electrical connection
                for dest in self.pops_and_nrns[dest_pop]:
                    for source in self.pops_and_nrns[source_pop]:
                        g_syn = g_max / len(self.pops_and_nrns[source_pop])
                        if self.network.connections[syn]['params']['rectified']:  # rectified
                            self.incoming_synapses[dest].append([source, False, True, g_syn, True, source, dest])
                            self.incoming_synapses[source].append([dest, False, True, g_syn, True, source, dest])
                        else:
                            self.incoming_synapses[dest].append([source, False, True, g_syn, False])
                            self.incoming_synapses[source].append([dest, False, True, g_syn, False])
            else:   # chemical connection
                if self.network.connections[syn]['params']['spiking']:
                    tau_s = self.network.connections[syn]['params']['synapticTimeConstant']
                    if self.delay:
                        delay = self.network.connections[syn]['params']['synapticTransmissionDelay']
                    for dest in self.pops_and_nrns[dest_pop]:
                        for source in self.pops_and_nrns[source_pop]:
                            g_syn = g_max / len(self.pops_and_nrns[source_pop])
                            if self.delay:
                                buffer = np.zeros(delay+1)
                                self.incoming_synapses[dest].append([source, True, False, g_syn, del_e, 0, self.dt/tau_s,buffer])
                            else:
                                self.incoming_synapses[dest].append([source, True, False, g_syn, del_e, 0, self.dt / tau_s])
                else:
                    for dest in self.pops_and_nrns[dest_pop]:
                        for source in self.pops_and_nrns[source_pop]:
                            g_syn = g_max / len(self.pops_and_nrns[source_pop])
                            self.incoming_synapses[dest].append([source,False, False, g_syn,del_e,0])

    def __calculate_time_factors__(self) -> None:
        """
        Precompute the time factors for the membrane voltage, firing threshold, and spiking synapses.
        :return: None
        """
        self.time_factor_membrane = self.dt / (self.c_m/self.g_m)
        if self.spiking:
            self.time_factor_threshold = self.dt / self.tau_theta

    def __initialize_propagation_delay__(self) -> None:
        """
        Create a buffer sized to store enough spike data for the longest synaptic propagation delay.
        :return: None
        """
        pass

    def __set_outputs__(self) -> None:
        """
        Build the output connectivity matrices for voltage and spike monitors and apply linear maps. Generate separate
        output monitors for each neuron in a population.
        :return: None
        """
        outputs = []
        index = 0
        for out in range(len(self.network.outputs)):
            source_pop = self.network.outputs[out]['source']
            num_source_neurons = self.network.populations[source_pop]['number']
            outputs.append([])
            for num in range(num_source_neurons):
                outputs[out].append(index)
                index += 1
        self.num_outputs = index

        self.output_voltage_connectivity = np.zeros([self.num_outputs, self.num_neurons])  # initialize connectivity matrix
        if self.spiking:
            self.output_spike_connectivity = np.copy(self.output_voltage_connectivity)
        self.outputs = np.zeros(self.num_outputs)
        for out in range(len(self.network.outputs)):  # iterate over the connections in the network
            source_pop = self.network.outputs[out]['source']  # get the source
            for i in range(len(self.pops_and_nrns[source_pop])):
                if self.network.outputs[out]['spiking']:
                    self.output_spike_connectivity[outputs[out][i]][
                        self.pops_and_nrns[source_pop][i]] = 1.0  # set the weight in the correct source and destination
                    # self.out_linear[outputs[out][i]] = 1.0
                else:
                    self.output_voltage_connectivity[outputs[out][i]][
                        self.pops_and_nrns[source_pop][i]] = 1.0  # set the weight in the correct source and destination

    def __debug_print__(self) -> None:
        """
        Print the values for every vector/matrix which will be used in the forward computation.
        :return: None
        """
        print('Input Connectivity:')
        print(self.input_connectivity)
        print('Output Voltage Connectivity')
        print(self.output_voltage_connectivity)
        if self.spiking:
            print('Output Spike Connectivity:')
            print(self.output_spike_connectivity)
        print('u:')
        print(self.u)
        print('u_last:')
        print(self.u_last)
        if self.spiking:
            print('theta_0:')
            print(self.theta_0)
            print('ThetaLast:')
            print(self.theta_last)
            print('Theta')
            print(self.theta)

    def __forward_pass__(self, inputs) -> Any:
        self.u_last = np.copy(self.u)
        if self.spiking:
            self.theta_last = np.copy(self.theta)

        i_app = np.matmul(self.input_connectivity, inputs)  # Apply external current sources to their destinations

        for nrn in range(self.num_neurons):
            i_syn = 0
            for syn in range(len(self.incoming_synapses[nrn])):
                neuron_src = self.incoming_synapses[nrn][syn]
                if neuron_src[1]:  # if spiking
                    neuron_src[5] = neuron_src[5] * (1-neuron_src[6])
                    i_syn += neuron_src[5] * (neuron_src[4] - self.u_last[nrn])
                elif neuron_src[2]: # if electrical
                    if neuron_src[4]:   # if rectified
                        if self.u_last[neuron_src[5]] > self.u_last[neuron_src[6]]:
                            i_syn += neuron_src[3] * (self.u_last[neuron_src[0]] - self.u_last[nrn])
                    else:
                        i_syn += neuron_src[3] * (self.u_last[neuron_src[0]] - self.u_last[nrn])
                else:   # if chemical
                    neuron_src[5] = np.maximum(0, np.minimum(neuron_src[3] * self.u_last[neuron_src[0]] / self.R, neuron_src[3]))
                    i_syn += neuron_src[5] * (neuron_src[4] - self.u_last[nrn])

            self.u[nrn] = self.u_last[nrn] + self.time_factor_membrane[nrn] * (-self.g_m[nrn] * self.u_last[nrn] + self.i_b[nrn] + i_syn + i_app[nrn])  # Update membrane potential
            if self.spiking:
                self.theta[nrn] = self.theta_last[nrn] + self.time_factor_threshold[nrn] * (-self.theta_last[nrn] + self.theta_0[nrn] + self.m[nrn] * self.u_last[nrn])  # Update the firing thresholds
                self.spikes[nrn] = np.sign(np.minimum(0, self.theta[nrn] - self.u[nrn]))  # Compute which neurons have spiked
        if self.spiking:
            for nrn in range(self.num_neurons):
                if self.delay:
                    # New stuff with delay
                    for syn in range(len(self.incoming_synapses[nrn])):
                        neuron_src = self.incoming_synapses[nrn][syn]
                        if neuron_src[1]:  # if spiking
                            neuron_src[7] = np.roll(neuron_src[7], 1)   # Shift buffer entries down
                            neuron_src[7][0] = self.spikes[neuron_src[0]]    # Replace row 0 with the current spike data
                            neuron_src[5] = np.maximum(neuron_src[5], (-neuron_src[7][-1]) * neuron_src[3])  # Update the conductance of connections which spiked
                self.u[nrn] = self.u[nrn] * (self.spikes[nrn] + 1)  # Reset the membrane voltages of neurons which spiked
        self.outputs = np.matmul(self.output_voltage_connectivity, self.u)
        if self.spiking:
            self.outputs += np.matmul(self.output_spike_connectivity, -self.spikes)

        return self.outputs


# if self.electrical:
#     i_syn += (torch.sum(self.g_electrical.to_dense() * self.u_last, 1).to_sparse() -
#               (self.u_last * torch.sum(self.g_electrical.to_dense(), 1)).to_sparse())
# if self.electrical_rectified:
#     # create mask
#     mask = (self.u_last.reshape(-1, 1) - self.u_last).transpose(0, 1) > 0
#     masked_g = mask * self.g_rectified.to_dense()
#     diag_masked = masked_g + masked_g.transpose(0, 1) - torch.diag(masked_g.diagonal())
#     i_syn += torch.sum(diag_masked * self.u_last, 1).to_sparse() - (self.u_last * torch.sum(diag_masked, 1)).to_sparse()


from setuptools import setup, find_packages


def readme_file_contents():
    with open('README.rst') as f:
        data = f.read()
    return data

setup(
    name='sns_toolbox',
    version='1.1.1',
    packages=find_packages(),
    url='https://github.com/wnourse05/SNS-Toolbox',
    license='Apache v2.0',
    author='William Nourse',
    author_email='nourse@case.edu',
    description='Tools for designing Synthetic Nervous Systems, and simulating them on various software/hardware backends',
    long_description='Documentation is available on ReadTheDocs  at https://sns-toolbox.readthedocs.io/en/latest/index.html',
    setup_requires=['wheel'],
    python_requires='>=3.5',
    install_requires=['graphviz',
                      'torch',
                      'numpy>=1.17'],

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: GPU',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ]
)



"""
Functions which are used within the design subpackage, but may not be necessary for the end user.
"""

"""
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
IMPORTS
"""

import warnings

SETOFVALIDCOLORS = {'aliceblue', 'antiquewhite', 'aqua', 'aquamarine', 'azure', 'beige', 'bisque', 'black', 'blanchedalmond',
          'blue', 'blueviolet', 'brown', 'burlywood', 'cadetblue', 'chartreuse', 'chocolate', 'coral',
          'cornflowerblue', 'cornsilk', 'crimson', 'cyan', 'darkblue', 'darkcyan', 'darkgoldenrod', 'darkgray',
          'darkgreen', 'darkgrey', 'darkkhaki', 'darkmagenta', 'darkolivegreen', 'darkorange', 'darkorchid', 'darkred',
          'darksalmon', 'darkseagreen', 'darkslateblue', 'darkslategray', 'darkslategrey', 'darkturquoise', 'darkviolet',
          'deeppink', 'deepskyblue', 'dimgray', 'dimgrey', 'dodgerblue', 'firebrick', 'floralwhite', 'forestgreen',
          'fuchsia', 'gainsboro', 'ghostwhite', 'gold', 'goldenrod', 'gray', 'grey', 'green', 'greenyellow', 'honeydew',
          'hotpink', 'indianred', 'indigo',	'ivory', 'khaki', 'lavender', 'lavenderblush', 'lawngreen',	'lemonchiffon',
          'lightblue', 'lightcoral', 'lightcyan', 'lightgoldenrodyellow', 'lightgray', 'lightgreen', 'lightgrey',
          'lightpink', 'lightsalmon', 'lightseagreen', 'lightskyblue', 'lightslategray', 'lightslategrey',
          'lightsteelblue', 'lightyellow', 'lime', 'limegreen', 'linen', 'magenta', 'maroon', 'mediumaquamarine',
                    'mediumblue', 'mediumorchid', 'mediumpurple', 'mediumseagreen', 'mediumslateblue', 'mediumspringgreen',
                    'mediumturquoise', 'mediumvioletred', 'midnightblue', 'mintcream', 'mistyrose', 'moccasin', 'navajowhite',
                    'navy', 'oldlace', 'olive', 'olivedrab', 'orange', 'orangered', 'orchid', 'palegoldenrod', 'palegreen',
                    'paleturquoise', 'palevioletred', 'papayawhip', 'peachpuff', 'peru', 'pink', 'plum', 'powderblue', 'purple',
                    'red', 'rosybrown', 'royalblue', 'saddlebrown', 'salmon', 'sandybrown', 'seagreen', 'seashell', 'sienna',
                    'silver', 'skyblue', 'slateblue', 'slategray', 'slategrey', 'snow', 'springgreen', 'steelblue', 'tan', 'teal',
                    'thistle', 'tomato', 'turquoise', 'violet', 'wheat', 'white', 'whitesmoke', 'yellow', 'yellowgreen'}
"""
Set of standard SVG color names.
"""

COLORSWHITETEXT = {'black', 'blue', 'blueviolet', 'brown', 'darkblue', 'darkmagenta', 'darkolivegreen', 'darkred',
          'darkslateblue', 'darkslategray', 'darkslategrey', 'darkviolet', 'dimgray', 'dimgrey', 'indigo', 'maroon',
                   'mediumblue', 'midnightblue', 'navy', 'olive', 'purple', 'saddlebrown', 'teal'}
"""
Set of background colors which work better with white text for improved visibility.
"""


def valid_color(color: str = None) -> bool:
    """
    Check if a given color is within the standard svg set.

    :param color: Desired color, default is None.
    :type color: str, optional
    :return: True if it is a valid color, false if not.
    :rtype: bool
    """
    if isinstance(color, str):
        if color in SETOFVALIDCOLORS:
            return True
        else:
            return False
    else:
        return False

def set_text_color(color: str = None) -> str:
    """
    Set the text color to white or black given a valid color string.

    :param color: Background color, default is None.
    :type color: str, optional
    :return: Text color based on background color.
    :rtype: str
    """
    if isinstance(color, str):
        if color in COLORSWHITETEXT:
            return 'white'
        else:
            return 'black'
    else:
        warnings.warn('Specified background color is not a string. Defaulting to black text color')
        return 'black'
